package tools;
import java.util.Enumeration;
import java.io.Serializable;
import random.*;
import tree.LblTree;


public class TreeGenerator implements Serializable {
	
	RandomInt depth;
	RandomInt fanout;
	RandomVal labels;
	RandomBoolean makeLeaf;
	
	double labelsToNodes; 
	int labelCnt;
	
	public TreeGenerator(RandomInt depth, RandomInt fanout, 
			RandomVal labels, RandomBoolean makeLeaf,
			double labelsToNodes) {
		this.depth = depth;
		this.fanout = fanout;
		this.labels = labels;
		this.makeLeaf = makeLeaf;
		this.labelsToNodes = labelsToNodes;
		this.labelCnt = -1;
	}
	
	
	public TreeGenerator(RandomInt depth, RandomInt fanout, 
			RandomVal labels, RandomBoolean makeLeaf,
			int labelCnt) {
		this.depth = depth;
		this.fanout = fanout;
		this.labels = labels;
		this.makeLeaf = makeLeaf;
		this.labelsToNodes = -1;
		this.labelCnt = labelCnt;
	}
	
	public void resetSeed(long seed) {
		depth = new RandomInt(depth.getMin(), depth.getMax(), 
				depth.getDistribution(), seed);
		fanout = new RandomInt(fanout.getMin(), fanout.getMax(), 
				fanout.getDistribution(), seed);
		labels = new RandomVal(labels.getDistribution(), seed);
		makeLeaf = new RandomBoolean(makeLeaf.getTrueProbability(), seed);
	}
	
	public LblTree getTree(int treeID) {	
		int totalDepth = depth.getInt();
		LblTree tree = getTree(totalDepth, totalDepth);
		tree.setTreeID(treeID);
		
		int numberOfLabels;
		if (labelsToNodes >= 0) {
			numberOfLabels = Math.max((int)Math.round(labelsToNodes * tree.getNodeCount()), 1);
		} else {
			numberOfLabels = labelCnt;
		}
		setRandomLabels(tree, numberOfLabels);
		return tree;
	}
	
	/**
	 * recursive method used by getTree(int)
	 */
	private LblTree getTree(int depthToLeaf, int totalDepth) {
		LblTree tree = new LblTree("", -1);
		int randomFanout = fanout.getInt();	
		if (depthToLeaf > 0) {
			boolean hasChildren = false;
			do {
				tree.removeAllChildren();
				for (int f = 0; f < randomFanout; f++) {
					if (makeLeaf.getBoolean()) {
						tree.add(getTree(0, totalDepth));
					} else {
						hasChildren = true;
						tree.add(getTree(depthToLeaf - 1, totalDepth));
					}
				}
			} while ((!hasChildren) && (randomFanout != 0));
		}
		return tree;
	}
	
	public void setRandomLabels(LblTree tree, int labelCount) {
		tree.setLabel(labels.getString(labelCount));
		for (Enumeration e = tree.children(); e.hasMoreElements();) {
			setRandomLabels((LblTree)e.nextElement(), labelCount);
		}	
	}
	
	public String toString() {
		String res = "";
		res += "  depth: " + depth + "\n";
		res += "  fanout: " + fanout + "\n";
		res += "  labels: " + labels + "\n";
		if (labelsToNodes >= 0) {
			res += "  (number of labels) / (number of nodes): " + 
			labelsToNodes + "\n";
		} else {
			res += "  number of labels: " + labelCnt + "\n";
		}
		res += "  makeLeaf: " + makeLeaf + "\n";
		return res;
	}
	
	public static void main(String[] args) {
		if (args.length != 6) {
			System.out.println("TreeGenerator d_min d_max f_min f_max l2n leafProbability");
		} else {
			int depthMin = Integer.parseInt(args[0]);
			int depthMax = Integer.parseInt(args[1]);
			int fanoutMin = Integer.parseInt(args[2]);
			int fanoutMax = Integer.parseInt(args[3]);
			double labelToNodes = Double.parseDouble(args[4]);
			double leafProb = Double.parseDouble(args[5]);
			TreeGenerator tg = 
				new TreeGenerator(new RandomInt(depthMin, depthMax),
						new RandomInt(fanoutMin, fanoutMax),
						new RandomVal(),
						new RandomBoolean(leafProb),
						labelToNodes);
			
			System.out.println(tg);
			LblTree t = tg.getTree(0);
			t.prettyPrint();
			System.out.println("Number of nodes: " + t.getNodeCount());
			System.out.println("Number of leaves: " + t.getLeafCount());
			System.out.println("Depth: " + t.getDepth());
			System.err.println(t);
			
		}
	}
}
