/*
 * Created on Feb 18, 2005
 */
package utility;

import hash.PrefixHash;

import java.io.StringWriter;

import xml.SvenPQGrams;
import distance.ButtomUpDist;
import distance.EditDist;
import distance.FanoutWeighting;
import distance.ProfileDist;
import distance.SvenDist;
import distance.SvenDistFromProfile;
import distance.TreeEmbedDist;
import distance.PQGramDist;
import distance.TreeDist;
import distance.HistoDist;
import distance.MergeDist;
import distance.BinaryBranchDist;
import distance.WeightedEditDist;

/**
 * Some static methods to evaluate command line parameters...
 * 
 * @author augsten
 */
final public class StringToTreeDist {

	public static boolean DEBUG = true;
	
	/**
	 * Only static methods...
	 */
	private StringToTreeDist() {
		super();
	}
	
	public static String getSyntax() {
		String s = 
			"'e': edit distance\n" +
			"'fweX': fanout weighted edit distance with leaf weight X\n" +
			"'b': bottom-up distance\n" +
			"'gk': edit distance embedding (Garofalakis & Kumar, PODS 2003)\n" +
			"'pq': pq-gram distance (default values for p and q)\n" +
			"'pXqY: pq-gram distance (p = X, q = Y)\n" +
			"'h': histogram distance (1,0-gram distance)\n" +
			"'m': merge distance\n" +
			"'bb': binary branch distance (Yang et al., SIGMOD 2005)\n" + 
			"'wpq': weighted pq-gram distance\n" +
			"'svenProfile': 2,3-grams with cross parsing\n" +
			"'svenPQGrams': 2,3-gram with Svens version\n" +
			"==> prefixing 'N' means 'normalize' (default for profile distances: 'bag norm')\n" +
			"==> prefixing 'D' means 'normalize with dice norm' (only for profile distance)\n"
			;	
		return s;
	}

	/**
	 * Depending on the argument, a different specialisation of a TreeDist Object is returned.
	 * <ul>
	 * </ul>
	 * @param strDist string argument from command line 
	 * @return TreeDist of the type corresponding to the argument, or null, if no match...
	 */
	public static TreeDist getTreeDist(String strDist) {
		TreeDist dist = null;
		boolean normalized = false;
		int typeOfNormalization = ProfileDist.BAG_NORM;
		if (strDist.charAt(0) == 'N') {
			normalized = true;
			strDist = strDist.substring(1);
		} else if (strDist.charAt(0) == 'D') {
			normalized = true;
			typeOfNormalization = ProfileDist.DICE_NORM;
			strDist = strDist.substring(1);
		}		
		
		if (strDist.equals("e")) {
			dist = new EditDist(normalized);			
		} else if (strDist.substring(0, Math.min(strDist.length(), 3)).equals("fwe")) {
			int leafWeight = Integer.parseInt(strDist.substring(3));
			dist = new WeightedEditDist(new FanoutWeighting(leafWeight), normalized);
		} else if (strDist.equals("b")) {
			dist = new ButtomUpDist(normalized);
		} else if (strDist.equals("m")) {
			dist = new MergeDist(normalized);
		} else if (strDist.equals("gk")) {
			dist = new TreeEmbedDist(normalized);
			((ProfileDist)dist).setTypeOfNormalization(typeOfNormalization);
		} else if (strDist.equals("pq")) {
			dist = new PQGramDist(2, 3, null, normalized);
			((ProfileDist)dist).setTypeOfNormalization(typeOfNormalization);
		} else if (strDist.equals("h")) {
			dist = new HistoDist(normalized);
			((ProfileDist)dist).setTypeOfNormalization(typeOfNormalization);
		} else if (strDist.equals("bb")) {
			dist = new BinaryBranchDist(normalized);
			((ProfileDist)dist).setTypeOfNormalization(typeOfNormalization);
		} else if (strDist.equals("wpq")) {
			dist = new distance.WeightedPQGramDist(1, 2, normalized);
			((ProfileDist)dist).setTypeOfNormalization(typeOfNormalization);
		} else if (strDist.equals("svenPQGrams")) {
			StringWriter sw1 = new StringWriter();
			StringWriter sw2 = new StringWriter();
			dist = new SvenDist(new SvenPQGrams(sw1, 2, 3),	
					new SvenPQGrams(sw2, 2, 3), normalized); 
		} else if (strDist.equals("svenProfile")) {
			dist = new SvenDistFromProfile(2, 3, normalized); 
		} else if (strDist.charAt(0) == 'p') {
			int qPos = strDist.indexOf('q');
			if ((qPos >= 2) && (qPos < strDist.length())) {
				String pStr = strDist.substring(1, qPos);
				String qStr = strDist.substring(qPos + 1, strDist.length());
				int p = Integer.parseInt(pStr);
				int q = Integer.parseInt(qStr);
				dist = new PQGramDist(p, q, null, normalized);
				((ProfileDist)dist).setTypeOfNormalization(typeOfNormalization);
			}
		}
		if (dist != null) {
			System.out.println("# distance: " + dist);			
		}
		return dist;
	}
	
	public static void main(String[] args) {
		if (args.length != 1) {
			System.out.println("Please give a parameter:\n" + 
					getSyntax());			
		} else {
			TreeDist dist = getTreeDist(args[0]);
			if (dist == null) {
				System.out.println("Unknown distance parameter. Please use on of the following:\n" + 
						getSyntax());									
			} else {
				System.out.println(dist);
			}
		}
	}
}
