package distance;

import java.io.FileReader;
import java.io.FileWriter;
import java.io.LineNumberReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.TreeSet;
import java.util.Arrays;
import java.util.Vector;
import java.sql.*;

import join.Intersect;

import tree.MMForest;
import util.FormatUtilities;

import distance.TreeDist;


public class EvaluatedDistMatrix {
	
	public static final double NO_VALUE = Double.MAX_VALUE;
	
	private double[][] distMatrix;
	
	private String[] rowLabels;
	private String[] colLabels;
	
	private int commonNames;
	
	private boolean[] correctMatches;
	
	private Hashtable labelIndices;
	
	private boolean evaluated;
	
	private int guessMatch;
	private int guessNomatch;
	private int correctMatch;
	private int incorrectMatch;
	private int matchedNomatch;
	private int matchNotFound;
	private int nomatchFound;
	private double recall;
	private double precision;
	
	/**
	 * @return Returns the correctMatch.
	 */
	public int getCorrectMatch() {
		return correctMatch;
	}
	/**
	 * @return Returns the guessMatch.
	 */
	public int getGuessMatch() {
		return guessMatch;
	}
	/**
	 * @return Returns the guessNomatch.
	 */
	public int getGuessNomatch() {
		return guessNomatch;
	}
	/**
	 * @return Returns the incorrectMatch.
	 */
	public int getIncorrectMatch() {
		return incorrectMatch;
	}
	/**
	 * @return Returns the matchedNomatch.
	 */
	public int getMatchedNomatch() {
		return matchedNomatch;
	}
	/**
	 * @return Returns the matchNotFound.
	 */
	public int getMatchNotFound() {
		return matchNotFound;
	}
	/**
	 * @return Returns the nomatchFound.
	 */
	public int getNomatchFound() {
		return nomatchFound;
	}
	
	public double getPrecision() {
		return precision;
	}
	public double getRecall() {
		return recall;
	}
	private void initLabelIndices() {
		// initialize labelIndices
		labelIndices = new Hashtable();
		for (int i = 0; i < rowLabels.length; i++) {
			labelIndices.put(rowLabels[i], new Integer(i));
		}
		for (int i = commonNames; i < colLabels.length; i++) {
			labelIndices.put(colLabels[i], new Integer(i));
		}		
	}
	
	private void initDistMatrix(double val) {
		// initialize dist matrix
		distMatrix = new double[rowLabels.length][colLabels.length];	  
		for (int i = 0; i < rowLabels.length; i++) {
			for (int j = 0; j < colLabels.length; j++) {
				distMatrix[i][j] = val;
			}
		}
		evaluated = false;
		correctMatches = new boolean[commonNames]; // by default all false!
	}
	
	/**
	 * <p>Reads a distance matrix from a file. The format in the file is:</p>
	 * 
	 * <p>rowLabel | colLabel | dist</p>
	 * 
	 * <p>meaning: the distance between the elements labeled rowLabel and
	 * colLabel are at distance dist.
	 * @param filename file containing the distance matrix
	 * @param seperator '|' in our example
	 * @throws IOException
	 */
	public EvaluatedDistMatrix(String filename, char seperator) throws IOException {
		
		TreeSet rowLabelsSet = new TreeSet();
		TreeSet colLabelsSet = new TreeSet();
		
		String inline;
		
		// first read the row- and column-labels
		
		LineNumberReader in = 
			new LineNumberReader(new FileReader(filename));
		
		long numLines = 0;      // counts only non-blank lines
		while ((inline = in.readLine()) != null) {
			inline = inline.trim();
			
			// simply skip blank lines
			if (!inline.equals("")) { 
				String rl = FormatUtilities.getField(0, inline, seperator);
				String cl = FormatUtilities.getField(1, inline, seperator);
				if ((rl == null) || (cl == null)) {
					throw(new IOException("Format error in file '" + filename +
							"' containing EvaluatedDistMatrix: " + 
							"Missing field(s) in line " + 
							(in.getLineNumber() + 1) + "."));
				}
				rowLabelsSet.add(rl);
				colLabelsSet.add(cl);
				numLines++;
			}
		}
		
		in.close();
		
		// order row- and column-labels
		
		int rows = rowLabelsSet.size();
		int cols = colLabelsSet.size();
		
		if ((rows * cols != numLines) || (numLines == 0)) {
			throw(new IOException("Format error in file '" + filename +
					"' containing EvaluatedDistMatrix: " + "Wrong number of lines."));
		}
		
		// for simpler code, copy labels to an array
		rowLabels = new String[rows];
		colLabels = new String[cols];
		{
			int i = 0;
			for (Iterator it = rowLabelsSet.iterator(); it.hasNext();i++) {
				rowLabels[i] = (String)it.next();
			}
			i = 0;
			for (Iterator it = colLabelsSet.iterator(); it.hasNext();i++) {
				colLabels[i] = (String)it.next();
			}
		}
		
		commonNames = orderLabels(rowLabels, colLabels);
		
		initLabelIndices();
		initDistMatrix(EvaluatedDistMatrix.NO_VALUE);
		
		// now read the distances
		
		in = new LineNumberReader(new FileReader(filename));
		
		while ((inline = in.readLine()) != null) {
			inline = inline.trim();
			
			// simply skip blank lines
			if (!inline.equals("")) { 
				String rl = FormatUtilities.getField(0, inline, seperator);
				String cl = FormatUtilities.getField(1, inline, seperator);
				String distStr = FormatUtilities.getField(2, inline, seperator);
				if ((distStr == null) || (rl == null) || (cl == null)) {
					throw(new IOException("Format error in file '" + filename +
							"' containing EvaluatedDistMatrix: " + 
							"Missing field(s) in line " + 
							(in.getLineNumber() + 1) + "."));
				}
				double dist = Double.parseDouble(distStr);
				
				int rowID = getIndex(rl);
				int colID = getIndex(cl);
				
				if (distMatrix[rowID][colID] == NO_VALUE) {
					distMatrix[rowID][colID] = dist;
				} else {
					throw(new IOException("Format error in file '" + filename +
							"' containing EvaluatedDistMatrix: " + 
							"Distance assigned twice in line " + 
							(in.getLineNumber() + 1) + "."));
				}
			}
		}	
	}
	
	
	public EvaluatedDistMatrix(String[] rowLabels, String[] colLabels) {
		Arrays.sort(rowLabels);
		Arrays.sort(colLabels);
		commonNames = orderLabels(rowLabels, colLabels);
		this.rowLabels = rowLabels;
		this.colLabels = colLabels;
		initLabelIndices();
		initDistMatrix(NO_VALUE);
	}
	
	/**
	 * Write a file that can be read by {@link #EvaluatedDistMatrix(String, char)}
	 * 
	 * @param filename
	 * @param seperator
	 * @throws IOException
	 */
	public void toFile(String filename, char seperator) throws IOException {
		BufferedWriter out = 
			new BufferedWriter(new FileWriter(filename));
		
		for (int i = 0; i < distMatrix.length; i++) {
			for (int j = 0; j < distMatrix[i].length; j++) {
				String outline = getRowLabel(i) + seperator + getColLabel(j) + seperator + distMatrix[i][j] + "\n";
				out.write(outline, 0, outline.length());
			}
		}
		out.close();
	}
	
	
	/**
	 * stores a distmatrix to a table. Assumes, that distances are INT values!
	 */
	public void toDB(Connection con, String table) throws SQLException {
		Statement s = con.createStatement();
		try {
			s.execute("DROP TABLE `" + table + "`");
		} catch (SQLException e) {
			System.out.println("Did not need to drop table " + table + ".");
		}
		s.execute("CREATE TABLE `" + table + "` (" +
				"`treeID1` INT NOT NULL ," +
				"`treeID2` INT NOT NULL ," +
		"`insec` DOUBLE NOT NULL)");
		for (int i = 0; i < distMatrix.length; i++) {
			StringBuffer qryInsert = 
				new StringBuffer("INSERT INTO `" + table + "` " + 
				"(`treeID1`,`treeID2`,`insec`) VALUES");		
			boolean inserts = false;
			for (int j = 0; j < distMatrix[i].length; j++) {
				if (distMatrix[i][j] != NO_VALUE) {
					inserts = true;
					qryInsert.append("(" + getRowLabel(i) + "," + 
							getColLabel(j) + "," + 
							distMatrix[i][j] + "),");
				}
			}
			if (inserts) {
				s.executeUpdate(qryInsert.substring(0, qryInsert.length()-1));
			}
		}
	}

	/**
	 * Read dist-matrix from database table of type Intersect.
	 * 
	 */
	public static EvaluatedDistMatrix fromDB(Intersect is) throws SQLException {
		
		ResultSet rs = 
			is.getStatement().executeQuery("SELECT DISTINCT " + is.getAtbTreeID1() + 
					" FROM " + is.getTblName());
		Vector v = new Vector();
		while (rs.next()) {
			v.add(rs.getString(is.getAtbTreeID1()));
		}
		String[] rowLabels = new String[v.size()];
		v.toArray(rowLabels);
		
		rs = is.getStatement().executeQuery("SELECT DISTINCT " + is.getAtbTreeID1() + 
				" FROM " + is.getTblName());
		v = new Vector();
		while (rs.next()) {
			v.add(rs.getString(is.getAtbTreeID1()));
		}
		String[] colLabels = new String[v.size()];
		v.toArray(colLabels);
		
		EvaluatedDistMatrix dm = new EvaluatedDistMatrix(rowLabels, colLabels);
		dm.initDistMatrix(1);
		
		rs = is.getStatement().executeQuery("SELECT * FROM " +  is.getTblName());
		while (rs.next()) {
			String i = rs.getString(is.getAtbTreeID1());
			String j = rs.getString(is.getAtbTreeID2());
			dm.setValue(i, j, rs.getDouble(is.getAtbDist()));
		}
		return dm;
	}

	
	/**
	 * read dist-matrix from a table that looks like table(treeID1, treeID2, dist).
	 * Attribute names do not matter, only positions (1,2,3)
	 */
	public static EvaluatedDistMatrix fromDB(Connection con, String tblName, 
			String treeID1, String treeID2, String dist) throws SQLException {
		
		Statement s = con.createStatement();
		
		ResultSet rs = s.executeQuery("SELECT DISTINCT " + treeID1 + " FROM " + tblName);
		Vector v = new Vector();
		while (rs.next()) {
			v.add(rs.getString(treeID1));
		}
		String[] rowLabels = new String[v.size()];
		v.toArray(rowLabels);
		
		rs = s.executeQuery("SELECT DISTINCT " + treeID2 + " FROM " + tblName);
		v = new Vector();
		while (rs.next()) {
			v.add(rs.getString(treeID2));
		}
		String[] colLabels = new String[v.size()];
		v.toArray(colLabels);
		
		EvaluatedDistMatrix dm = new EvaluatedDistMatrix(rowLabels, colLabels);
		
		rs = s.executeQuery("SELECT * FROM " +  tblName);
		while (rs.next()) {
			String i = rs.getString(treeID1);
			String j = rs.getString(treeID2);
			dm.setValue(i, j, rs.getDouble(dist));
		}
		return dm;
	}
	
	/** 
	 * rowLabels and colLabels MUST be sorted!!
	 */
	public static int orderLabels(String[] rowLabels, String[] colLabels) {
		int rows = rowLabels.length;
		int cols = colLabels.length;
		
		// make temporary copies of the original arrays
		String[] tmpRowLabels = new String[rows];
		String[] tmpColLabels = new String[cols];
		
		for (int i = 0; i < rows; i++) {
			tmpRowLabels[i] = rowLabels[i];
		}
		for (int i = 0; i < cols; i++) {
			tmpColLabels[i] = colLabels[i];
		}
		
		// order the labels
		int commonNames = 0;      // begin of arrays rowLabels and colLabels
		int endRows = rows - 1; // end of rowLabels
		int endCols = cols - 1; // end of colLabels
		
		int r = 0;
		int c = 0;
		
		while ((r < rows) && (c < cols)) {
			String nameRow = tmpRowLabels[r];
			String nameCol = tmpColLabels[c];
			int cmp = nameRow.compareTo(nameCol);
			if (cmp == 0) {   // nameRow == nameCol
				rowLabels[commonNames] = nameRow;
				r++;
				colLabels[commonNames] = nameCol;
				c++;
				commonNames++;
			} else if (cmp < 0) {  // nameRow < nameCol
				rowLabels[endRows] = nameRow;
				r++;
				endRows--;
			} else {               // nameRow > nameCol
				colLabels[endCols] = nameCol;
				c++;
				endCols--;
			}
		}
		// copy remaining labels
		for (int i = r; i < rows; i++) {
			rowLabels[endRows] = tmpRowLabels[i];
			endRows--;
		}
		for (int i = c; i < cols; i++) {
			colLabels[endCols] = tmpColLabels[i];
			endCols--;
		}
		return commonNames;
	}
	

	
	public int minXY(int row) {
		double min = Double.MAX_VALUE;
		int minCnt = -1;
		int minPos = -1;
		for (int j = 0; j < distMatrix[row].length; j++) {
			if (min > distMatrix[row][j]) {
				min = distMatrix[row][j];
				minPos = j;
				minCnt = 1;
			} else if (min == distMatrix[row][j]) {
				minCnt++;
			}
		}
		if (minCnt == 1) {
			for (int i = 0; i < distMatrix.length; i++) {
				if (min >= distMatrix[i][minPos]) {
					if (i != row) {
						return -1;
					}
				}
			}
			return minPos;
		} else {
			return -1;
		}
	}
	
	public int getCommonNames() {
		return commonNames;
	}
	
	public void evaluateDistMatrix() {
		evaluated = true;
		int k = commonNames;
		guessMatch = 0;
		guessNomatch = 0;
		correctMatch = 0;
		incorrectMatch = 0;
		matchedNomatch = 0;
		matchNotFound = 0;
		nomatchFound = 0;
		for (int i = 0; i < distMatrix.length; i++) {
			int pos = minXY(i);
			if (i < k) {    // initialize
				correctMatches[i] = false;
			}
			if (pos == -1) {
				guessNomatch++;
				if (i < k) {
					matchNotFound++;
				} else {
					nomatchFound++;
				}
			} else {
				guessMatch++;
				if (i < k) {
					if (pos == i) {
						correctMatch++;
						correctMatches[i] = true;
					} else {
						incorrectMatch++;
					}
				} else {
					matchedNomatch++;
				}
			}
		}
		precision = correctMatch / (double)guessMatch;
		recall = correctMatch / (double)getCommonNames();
	}
	
	public static final int LATEX = 0;
	public static final int SCREEN = 1;
	
	public void printEvaluation (int style) {
		if (!evaluated) {
			evaluateDistMatrix();
		}
		switch (style) {
		case LATEX:
			System.out.print("{" + guessMatch + "}");
			System.out.print("{" + guessNomatch + "}");
			System.out.print("{" + correctMatch + "}");
			System.out.print("{" + incorrectMatch + "}");
			System.out.print("{" + matchedNomatch + "}");
			System.out.print("{" + matchNotFound + "}");
			System.out.print("{" + nomatchFound + "}");
			break;
		case SCREEN:
			System.out.println("Number of possible matches: " + getCommonNames());
			System.out.println("guessMatch: "+ guessMatch);
			System.out.println("guessNomatch: "+ guessNomatch);
			System.out.println("  correctMatch: "+ correctMatch + " (" + 
					((double)correctMatch / getCommonNames() * 100) + ")");
			System.out.println("  incorrectMatch: "+ incorrectMatch);
			System.out.println("  matchedNomatch: "+ matchedNomatch);
			System.out.println("  matchNotFound: "+ matchNotFound);
			System.out.println("  nomatchFound: "+ nomatchFound);
			System.out.println("  ----\n  precision = " + precision);
			System.out.println("  recall = " + recall);
			System.out.println("  f-measure = " + (2 * precision * recall / (precision+recall)));
			break;
		}
	}
	
	private int checkRow(int r, double tau) {
		int posHit = -1;
		for (int c = 0; c < this.getColNum(); c++) {
			if (this.distMatrix[r][c] <= tau) {
				if (posHit == -1) { //first hit
					posHit = c;
				} else {
					return -1;
				}
			}
		}
		if (posHit == -1) { // no match
			return -1;
		}
		if (!this.getColLabel(posHit).equals(this.getRowLabel(r))) {
			return -1;
		}
		return posHit;
	}
	
	private boolean checkCol(int colHit, double tau) {
		for (int r = 0; r < this.getRowNum(); r++) {
			if ((r != colHit) && (this.distMatrix[r][colHit] <= tau)) {
				return false;
			} 
		}
		return true;
	}
	
	/** 
	 * Count correct and incorrect matches for a join with threshold tau.
	 * @param tau
	 */
	public int evaluateUnique(double tau) {
		int hits = 0;
		for (int r = 0; r < this.getRowNum(); r++) {
			int colHit = checkRow(r, tau);
			if ((colHit >= 0) && (checkCol(colHit, tau))) {
				hits++;
			}
		}
		return hits;
	}
	
	public int evaluate(double tau, boolean correct) {
		int cor = 0;
		int incor = 0;
		for (int r = 0; r < this.getRowNum(); r++) {
			for (int c = 0; c < this.getColNum(); c++) {
				if (this.distMatrix[r][c] <= tau) {
					if (this.getColLabel(c).equals(this.getRowLabel(r))) {
						cor++;
					} else {
						incor++;
					}
				}
			}
		}
		if (correct) {
			return cor;
		} else {
			return incor;
		}
	}
	
	public int getIndex(String label) {
		Integer i = (Integer)labelIndices.get(label);
		if (i != null) {
			return i.intValue();
		} else {
			return -1;
		}	
	}
	
	public String getRowLabel(int row) {
		if (row < rowLabels.length) {
			return rowLabels[row];
		} else {
			return null;
		}
	}
	
	public String getColLabel(int col) {
		if (col < colLabels.length) {
			return colLabels[col];
		} else {
			return null;
		}
	}
	
	public void setValue(int row, int col, double dist) {
		distMatrix[row][col] = dist;
		evaluated = false;
	}
	
	public double getValue(int row, int col) {
		return distMatrix[row][col];
	}
	
	public int getRowNum() {
		return rowLabels.length;
	}
	
	public int getColNum() {
		return colLabels.length;
	}
	
	public void setValue(String rowLabel, String colLabel, double dist) {
		distMatrix[getIndex(rowLabel)][getIndex(colLabel)] = dist;
		evaluated = false;
	}
	
	public double getValue(String rowLabel, String colLabel) {
		return distMatrix[getIndex(rowLabel)][getIndex(colLabel)];
	}
	
	public void prettyPrint() {
		prettyPrint(5, 0, rowLabels.length - 1, 0, colLabels.length - 1);
	}
	
	public void prettyPrint(int tabSize, int fromRow, int toRow, int fromCol, int toCol) {
		System.out.print(FormatUtilities.resizeEnd("", tabSize) + " | ");
		for (int j = fromCol; j <= toCol; j++) {
			System.out.print(FormatUtilities.resizeFront(colLabels[j], tabSize) + " ");
		}
		System.out.print("\n" + FormatUtilities.resizeEnd("", tabSize, '-') + " | ");
		System.out.println(FormatUtilities.resizeEnd("", (tabSize + 1) * (toCol - fromCol + 1), '-'));
		for (int i = fromRow; i <= toRow; i++) {
			System.out.print(FormatUtilities.resizeFront(rowLabels[i], tabSize) + " | ");
			for (int j = fromCol; j <= toCol; j++) {
				System.out.print(FormatUtilities.resizeFront(distMatrix[i][j] + "", tabSize) + " ");
			}
			System.out.println();
		}
		System.out.print(FormatUtilities.resizeEnd("", tabSize, '-') + " | ");
		System.out.println(FormatUtilities.resizeEnd("", (tabSize + 1) * (toCol - fromCol + 1), '-'));
	}
	
	public boolean[] getCorrectMatches() {
		if (!evaluated) {
			evaluateDistMatrix();
		}	
		return correctMatches;
	}
	
	/**
	 * Give a list of labels for correct matches.
	 */
	public String[] getCorrectMatchingLabels() {
		if (!evaluated) {
			evaluateDistMatrix();
		}	
		Vector cmv = new Vector(commonNames);
		for (int i = 0; i < commonNames; i++) {
			if (correctMatches[i]) {
				cmv.add(getRowLabel(i));   // for correct Matches row- and col-names are equal!
			}
		}
		String[] cma = new String[cmv.size()];
		return (String[])cmv.toArray(cma);
	}
	
	/**
	 *  Creates a EvaluatedDistMatrix of size |forest1| x |forest2|, using the distance metrics dist, so that
	 *
	 *  dm[i][j] = dist.treeDist(forest1[i], forest2[j])
	 *
	 *  Caution: The arrays storing the trees in this class have different order than the treenames in the
	 *  distMatrix. Therefore you can access the elements in the distMatrix only by Treenames.
	 *  This is probably a source of inefficiency and should be changed!! You don't feel this inefficiency very strong
	 *  because the number of trees is normally low and the cost for EvaluatedDistMatrix.getIndex() is normally much lower 
	 *  then pq-gram-distance, althoug it is log(n) (instead of constant, as array access should be!!).
	 *  Difference for 300x300 trees: 100ms
	 *  
	 *  @param forest1 first forest
	 *  @param forest2 second forest
	 *  @param dist distance metrics (pq-gram-dist, edit-dist, buttom-up-dist etc.)
	 *
	 */
	public EvaluatedDistMatrix(MMForest forest1, MMForest forest2, TreeDist dist) {
		// construct rowLabels and colLabels
		String[] rowLabels = new String[forest1.size()];
		String[] colLabels = new String[forest2.size()];
		for (int i = 0; i < forest1.size(); i++) {
			rowLabels[i] = forest1.getTreeAt(i).getTreeID() + "";
		}
		for (int i = 0; i < forest2.size(); i++) {
			colLabels[i] = forest2.getTreeAt(i).getTreeID() + "";
		}
		// the same as constructor {@link EvaluatedDistMatrix(String[], String[])}
		Arrays.sort(rowLabels);
		Arrays.sort(colLabels);
		commonNames = orderLabels(rowLabels, colLabels);
		this.rowLabels = rowLabels;
		this.colLabels = colLabels;
		initLabelIndices();
		initDistMatrix(EvaluatedDistMatrix.NO_VALUE);
		// fill EvaluatedDistMatrix
		for (int i = 0; i < forest1.size(); i++) {
			System.out.println(i + "/" + forest1.size());
			for (int j = 0; j < forest2.size(); j++) {	
				System.out.println(i + "/" + forest1.size() + " <--> " +  j + "/" + forest2.size() );
				double d;
				d = dist.treeDist(forest1.getTreeAt(i), forest2.getTreeAt(j));
				setValue(forest1.getTreeAt(i).getTreeID() + "", 
						forest2.getTreeAt(j).getTreeID() + "", d);
			}
		}
	}
	/**
	 * Give a list of labels that where matched correctly in this distmatrix, but not in the argument
	 * dist matrix. Returns NULL, if number of common names do not match. Labels of this distmatrix are used. 
	 * It's NOT checked, whether they are equal to the ones in the other matrix.
	 */
	public String[] getBetterMatchingLabels(EvaluatedDistMatrix dm) {
		if (!evaluated) {
			evaluateDistMatrix();
		}	
		if (commonNames != dm.getCommonNames()) {
			return null;
		}
		Vector cmv = new Vector(commonNames);
		for (int i = 0; i < commonNames; i++) {
			if ((correctMatches[i]) && (!(dm.getCorrectMatches()[i]))) {		                           
				cmv.add(getRowLabel(i));   // for correct Matches row- and col-names are equal!
			}
		}
		String[] cma = new String[cmv.size()];
		return (String[])cmv.toArray(cma);
	}
		
	public static void main(String[] args) throws IOException { 
		EvaluatedDistMatrix dm1 = new EvaluatedDistMatrix(args[0], '|');
		EvaluatedDistMatrix dm2 = new EvaluatedDistMatrix(args[1], '|');
		
		dm1.evaluateDistMatrix();
		dm1.printEvaluation(EvaluatedDistMatrix.SCREEN);
		
		dm2.evaluateDistMatrix();
		dm2.printEvaluation(EvaluatedDistMatrix.SCREEN);
		
		String[] bm1 = dm1.getBetterMatchingLabels(dm2);
		
		// System.out.println(bm1.length + " better matches in " + args[0]);
		MMForest an = new MMForest("an.tbl");
		MMForest ae = new MMForest("ae.tbl");			       
		
		int[] intBm1 = new int[bm1.length];
		
		for (int i = 0; i < bm1.length; i++) {
			intBm1[i] = Integer.parseInt(bm1[i]);
		}
		
		MMForest an1 = an.getSubforest(intBm1);
		MMForest ae1 = ae.getSubforest(intBm1);
		
		for (int i = 0; i < an1.size(); i++) {
			System.out.print(an1.getTreeAt(i).getLabel() + " <-> ");
			System.out.println(ae1.getTreeAt(i).getLabel());
			
			//	    System.out.println(an1.getTreeAt(i).toLatex());
			//      System.out.println(ae1.getTreeAt(i).toLatex());
			//      System.out.println("\n\n");
		}
	}
	
	
}
