/*
 * Created on Feb 9, 2005
 */
package distance;

import tree.LblTree;

import java.util.Enumeration;
import java.util.LinkedList;
import java.util.ListIterator;
import parser.Splitter;

/**
 * Implementation of the distance measure proposed by
 * Minos Garofalakis, Amit Kumar, "Correlating XML Data Streams Using
 * Tree-Edit Distance Embeddings", PODS 2002. 
 *
 *  * @author augsten
 */
public class TreeEmbedDist extends ProfileDist {
	
	public static boolean DEBUG = false;
	
	private static final int HASH_CONST = 1000;		/* To be used in hash function */
	private static final int HASH_A = 7901;
	private static final int HASH_B = 7877;
	private static final int HASH_C = 6007;
	private static final int HASH_D = 6073;
	private static final int HL = 31;
	private static final int MOD = Integer.MAX_VALUE;
	
	public TreeEmbedDist(boolean normalized) {
		super(normalized);
	}	
	
	
			
	/* (non-Javadoc)
	 * @see distance.ProfileDist#createProfile(tree.LblTree)
	 */
	@Override
	public Profile createProfile(LblTree t) {
		Profile p = getParseVector(t);
		return p;
	}
	
	private static Profile getParseVector(LblTree t) {
		HashTree ht = new HashTree(t);
		VectorProfile parseVec = new VectorProfile(ht.getNodeCount() * 2, ht.getNodeCount()); 
		int phase = 0;
		do {
			if (DEBUG) {
				ht.prettyPrint();
				System.out.println(ht.toLatex());
			}
			for (Enumeration e = ht.breadthFirstEnumeration(); e.hasMoreElements();) {
				HashTree n = (HashTree)e.nextElement();
				parseVec.add(new SubtreePhaseCombi(n.getHashVal(), phase, n.getLabel()));
			}
			LinkedList[] l1 = getValidSubtrees(ht);
			handleLoneLeafs(l1[2], ht); 
			handleChains(l1[0], ht);
			handleLeafSeq(l1[1], ht);
			phase++; // next phase in parsing hierarchy
		} while (ht.getNodeCount() != 1);
		if (DEBUG) {
			ht.prettyPrint();
		}
		parseVec.add(new SubtreePhaseCombi(ht.getHashVal(), phase, ht.getLabel()));		
		if (DEBUG) {
			System.out.println("parse array (" + parseVec.size() + " elements):"); 
			for (int i = 0; i < parseVec.size(); i++) {
				System.out.print(parseVec.elementAt(i) + " ");
			}
			System.out.println();
		}
		return parseVec;
	}
	
	private static void handleChains(LinkedList l, HashTree t) {
		for (ListIterator it = l.listIterator(); it.hasNext();) {
			// one of the chains in the tree...	
			LinkedList list = (LinkedList)it.next();
			HashTree[] chain = (HashTree[])list.toArray(new HashTree[list.size()]);  
			
			// split the chain
			int[] split = split(chain);
			
			HashTree p = null;	// new node replacing previous block in chain

			// backup the subtree following the chain
			HashTree subtree = null;
			if (!chain[chain.length - 1].isLeaf()) {
				subtree = (HashTree)chain[chain.length - 1].getFirstChild(); // not leaf => first child exits
			}
			
			for (int s = 0; s < split.length; s++) {
				// length of block (2 or 3)
				int blocklength = chain.length - split[s];
				if (s < split.length - 1) {
					blocklength = split[s + 1] - split[s];
				}
				// store hash codes of the nodes in this block into contract
				int[] contract = new int[blocklength];
				for (int c = 0; c < contract.length; c++) {
					contract[c] = chain[split[s] + c].getHashVal();
					if (c == 0) {
						// we are starting a new block
						if (p != null) {
							// this is not the first block
							chain[split[s]].removeFromParent();
							p.add(chain[split[s]]);
						}				
						// remember the first node in the block (will get it's hashcode later...)
						p = chain[split[s]];
					} else {
						// just forget about the second and third node in a block
						chain[split[s] + c].removeFromParent();
					}
				}
				//chain[split[s]].setHashVal(hash(contract));
				// give p its new hash value
				p.setHashVal(hashChain(contract));
				// form new label
				if (DEBUG) {
					String label = "[";
					for (int c = 0; c < contract.length; c++) {
						label += chain[split[s] + c].getLabel();
						if (c < contract.length - 1) {
							label += ",";
						}
					}
					label += "]";
					chain[split[s]].setLabel(label);
				}
			}
			// link subtree after chain to last of new chain nodes
			if (subtree != null) {
				p.add(subtree);
			}
		}
	}
	
	private static int[] split(HashTree[] l) {
		// copy all hashvalues in an array and split them
		int[] intChain = new int[l.length];
		for (int i = 0; i < l.length; i++) {
			intChain[i] = l[i].getHashVal();
		}
		return Splitter.split(intChain);		
	}
	
	private static void handleLeafSeq(LinkedList l, HashTree t) {
		for (ListIterator it = l.listIterator(); it.hasNext();) {
			// one of the chains in the tree...	
			LinkedList list = (LinkedList)it.next();
			HashTree[] chain = (HashTree[])list.toArray(new HashTree[list.size()]);  
			
			// split the chain
			int[] split = split(chain);
			
			for (int s = 0; s < split.length; s++) {
				// length of block (2 or 3)
				int blocklength = chain.length - split[s];
				if (s < split.length - 1) {
					blocklength = split[s + 1] - split[s];
				}
				// store hash codes of the nodes in this block into contract
				int[] contract = new int[blocklength];
				for (int c = 0; c < contract.length; c++) {
					contract[c] = chain[split[s] + c].getHashVal();
					if (c != 0) {
						// get child-ID "by hand"
						HashTree child = chain[split[s] + c];
						HashTree par = (HashTree)(child.getParent());
						int index = 0;
						for (int chCnt = 0; chCnt < par.getChildCount(); chCnt++) {
							if (child == par.getChildAt(chCnt)) {
								index = chCnt;
								break;
							}
						}						
						par.remove(index);
					}
				}				
				chain[split[s]].setHashVal(hashLeaf(contract));
				// form new label
				if (DEBUG) {
					String label = "(";
					for (int c = 0; c < contract.length; c++) {
						label += chain[split[s] + c].getLabel();
						if (c < contract.length - 1) {
							label += ",";
						}
					}
					label += ")";
					chain[split[s]].setLabel(label);
				}
			}
		}
	}
	
	private static void handleLoneLeafs(LinkedList l, HashTree t) {
		for (ListIterator it = l.listIterator(); it.hasNext();) {
			HashTree n = (HashTree)it.next();
			HashTree p = (HashTree)n.getParent();
			n.removeFromParent();
			p.setHashVal(hashChain(new int[] {n.getHashVal(), p.getHashVal()}));
			if (DEBUG) {
				p.setLabel("{" + p.getLabel() + "-" + n.getLabel() + "}");
			}
		}
	}
	
	private static int hashLeaf(int[] block) {
		int result;
		int x = (block[0] * HASH_CONST * HASH_CONST);
		x += block[1] * HASH_CONST;
		if (block.length == 3) {
			x += block[2];
		}
		
		result = (HASH_A * x) + HASH_B;
		result = ((result >> HL) + result) & MOD;
		return (result);
	}
	
	private static int hashChain(int[] block) {
		int result;
		int x = (block[0] * HASH_CONST * HASH_CONST);
		x += block[1] * HASH_CONST;
		if (block.length == 3) {
			x += block[2];
		}
		
		result = (HASH_C * x) + HASH_D;
		result = ((result >> HL) + result) & MOD;
		return (result);
	}

	
	/**
	 * Get chains, leftmost lone leaf children and contiguous leaf-child subsequences.
	 * 
	 * @param t1
	 * @return {chainList, leafsList, loneLeafs}, where chainList is a list of chains (each chain is a LinkeList of nodes), 
	 * leafsList is a list of maximal conitguous leaf-child subsequenes (each sequence is a LinkedList of nodes), and loneLeafs
	 * is a list of leftmost lone leaf nodes.
	 */
	private static LinkedList[] getValidSubtrees(LblTree t1) {
		LinkedList chain = null;
		LinkedList chainList = new LinkedList();
		LinkedList leafs = null;
		LinkedList leafsList = new LinkedList();
		LinkedList loneLeaf = new LinkedList(); // leafmost lone leafs
		for (Enumeration e = t1.preorderEnumeration(); e.hasMoreElements();) {
			LblTree n = (LblTree)e.nextElement();
			if (n.isLeaf() && !n.isRoot()) {
				// n is a leaf 
				if (chain != null) {
					// if you are within a chain, then also n is in this chain and ends it
					chain.add(n);
					chainList.add(chain);
					chain = null;				
				} else {
					// do nothing 
				}
			} else if ((n.getChildCount() == 1)  && !n.isRoot()) {
				// this is a degree-2 node
				if (chain == null) {
					// you are not within a chain, but maybe at the start
					if ((n.getFirstChild().getChildCount() == 1) || (n.getFirstChild().isLeaf())) {
						// it is the start of a chain
						chain = new LinkedList();
						chain.add(n);
					} else {
						// do nothing
					}
				} else {  
					// you are already within a chain
					chain.add(n);
				}
			} else {
				// n has more than 1 child (or is the root node)

				// is there a chain to end?
				if (chain != null) {
					chainList.add(chain);
					chain = null;
				}				

				// check for leaf children
				boolean leftmost = true;
				for (int i = 0; i < n.getChildCount(); i++) {					
					LblTree c = (LblTree)n.getChildAt(i);
					if (c.isLeaf()) {
						// a leaf starts a new leaf sequence or
						// is added to an existing one...
						if (leafs == null) {
							leafs = new LinkedList();
						}
						leafs.add(c);
					} 
					if ((leafs != null) && (!c.isLeaf() || (i == n.getChildCount() - 1))) {
						// there is a leaf sequence to end
						if (leafs.size() == 1) {
							if (leftmost) {
								// leaf sequence to end is a leftmost lone leaf
								loneLeaf.add(leafs.getFirst());
								leafs = null;
								leftmost = false;
							} else {
								// leaf seq is lone leaf, but not leftmost
								// forget it
								leafs = null;
							}
						} else {
							// leaf sequence of more than 1 leafs
							leafsList.add(leafs);
							leafs = null;
						}
					}
				}
				if (leafs != null) {
					System.out.println("FEHLER!!!");
					System.exit(-1);				
				}
			}
		}
		debugPrint2DList("chain list", chainList);
		debugPrint2DList("leafs list", leafsList);
		debugPrintList("lone leafs", loneLeaf);		
		return new LinkedList[] {chainList, leafsList, loneLeaf};
	}
	
	
	
	private static void debugPrint2DList(String name, LinkedList ll) {
		if (DEBUG) {
		System.out.println(name + ": ");
		int i = 0;
		for (ListIterator it = ll.listIterator(); it.hasNext();) {
			LinkedList l = (LinkedList)it.next();
			debugPrintList(i + ": ", l);
			i++;
		}
		}
	}

	private static void debugPrintList(String name, LinkedList l) {
		if (DEBUG) {
		System.out.print(name + ": ");
		for (ListIterator it = l.listIterator(); it.hasNext();) {
			HashTree n = (HashTree)it.next();
			System.out.print(n.getLabel());
			if (it.hasNext()) {
				System.out.print(",");
			}
		}
		System.out.println();
		}
	}
	
	/*
	 * Questions:
	 * 1) What about leafs in chains? According to example, leaf is in chain, according to algo it is not...
	 * --> lets follow the example (this solves also problem 2)
	 * 2) What happens to a single leaf of a node? Is it a lone child? (according to definition: yes)
	 * --> if chains can also be leafs, then a single leaf node forms always a chain with its parent
	 * 3) todo: 
	 * --> probably its clever to save lone leaf children with there parents in a list, so you can deal with them
	 * the same way you deal with chains and leaf sequnces (when re-labeling them...). 
	 */
	
	public static void main(String[] args) {
		DEBUG = true;
		String treeStr1 = "1:{str{1{A{1}{2}{3}}{{1}{2}{3}}}{2{{1}{2}{3}}}{3}{4{A{1}{2}{3}{10}{12}}{B{1}{2}{3}{15}{20}}}}";
		//String treeStr1 = "1:{x{a}{a}{a}{a}{a}}";
		LblTree t1 = LblTree.fromString(treeStr1);
		String treeStr2 = "2:{x{a{a{a{a{a}}}}}}";
		LblTree t2 = LblTree.fromString(treeStr2);		
		double d = (new TreeEmbedDist(false)).treeDist(t1, t2);
		System.out.println("\nDISTANCE:\nd=" + d);
		d = new TreeEmbedDist(true).treeDist(t1, t2);
		System.out.println("\nDISTANCE:\nd=" + d);
	}
}

class HashTree extends LblTree {
	
	int hashVal;
	
	public HashTree(LblTree t) {
		super(t.getLabel(),t.getTreeID());
		setHashVal(t.getLabel().hashCode());
		for (int i = 0; i < t.getChildCount(); i++) {
			this.add(new HashTree((LblTree)t.getChildAt(i)));
		}
	}
	
	/**
	 * @return Returns the hashVal.
	 */
	public int getHashVal() {
		return hashVal;
	}
	/**
	 * @param hashVal The hashVal to set.
	 */
	public void setHashVal(int hashVal) {
		this.hashVal = hashVal;
	}
}

class SubtreePhaseCombi implements Comparable {
	
	private int	hashVal;
	private int phase;
	private String label = null; // optional
	
	private SubtreePhaseCombi(int hashVal, int phase) {
		this.hashVal = hashVal;
		this.phase = phase;
	}
	
	public SubtreePhaseCombi(int hashVal, int phase, String label) {
		this(hashVal, phase);
		this.label = label;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object arg0) {
		SubtreePhaseCombi spc = (SubtreePhaseCombi)arg0;
		if (spc.hashVal == hashVal) {
			return phase - spc.phase;
		} else {
			return hashVal - spc.hashVal; // negative, if hashVal < spc.hashVal, positive otherwise
		}
	}
	
}
