/*
 * Created on 10-Apr-06
 */
package intervalenc;

import hash.FixedLengthHash;
import hash.HashValue;
import hash.PrefixHash;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.LinkedList;

import join.PQGramTbl;
import join.ProfileSizeTbl;

import mysqltools.MySqlInsertBuffer;
import mysqltools.MySqlTools;

public class PQGramFactory {
	
	public final static int ORDERED = 0;
	public final static int UNORDERED = 1;
	
	private PQGramFactory() {
	}

	/**
	 * Computes pq-grams for forest, given in preorder.
	 * None of the tables is indexed. Index the profile size table before you use it.
	 * 
	 * @param preorder
	 * @param pqg
	 * @param ps profile size table
	 * @param p
	 * @param q
	 * @param type
	 * @throws Exception
	 */
	public static void getPQGrams(Cursor preorder, PQGramTbl pqg, ProfileSizeTbl ps, 
			int p, int q) throws Exception {
		// Start creating pq-grams
		preorder.next();
		while (!preorder.isAfterLast()) {
			IntervalEncNode root = preorder.fetchNode();
			int treeID = root.getTreeID();
			pqg.resetInsertCnt();
			getOrderedPQGrams(preorder, pqg, pqg.getHf().getEmptyRegister(p), q);
			ps.insertTree(treeID, pqg.getInsertCnt());
		}
		pqg.setPsTbl(ps);
		pqg.close();
		ps.close();
		
	}

	/**
	 * Computes pq-grams for forest, given in preorder.
	 * None of the tables is indexed. Index the profile size table before you use it.
	 * 
	 * @param preorder
	 * @param pqg
	 * @param ps profile size table
	 * @param p
	 * @param q
	 * @param w
	 * @throws Exception
	 */
	public static void getUnorderedPQGrams(Cursor preorder, PQGramTbl pqg, ProfileSizeTbl ps, 
			int p, int q, int w) throws Exception {
		// Start creating pq-grams
		preorder.next();
		while (!preorder.isAfterLast()) {
			IntervalEncNode root = preorder.fetchNode();
			int treeID = root.getTreeID();
			pqg.resetInsertCnt();
			getUnorderedPQGrams(preorder, pqg, pqg.getHf().getEmptyRegister(p), q, w);
			ps.insertTree(treeID, pqg.getInsertCnt());
		}
		pqg.setPsTbl(ps);
		pqg.close();
		ps.close();
		
	}
	
	
	/**
	 * Calculates all pq-grams of the subtree rooted in the node n at
	 * the current position in "preorder" It has some side-effects:
	 *
	 * <ul>
	 * 
	 * <li>cursor in preorder is set to the next node that is not a
	 * descendant of n (or to the root of the next tree, if n is the
	 * root node of a tree)
	 *
	 * </ul>
	 *
	 * @param anc next p ancestors of n
	 * @param sib next q left siblings of n
	 * @return next node in preorder
	 *
	 */
	private static IntervalEncNode getOrderedPQGrams(Cursor preorder, PQGramTbl pqg, LinkedList anc_in, int q) throws SQLException {
		
		// copy input shift register (so it will not be changed)
		LinkedList anc = (LinkedList)anc_in.clone();
		
		// get current node
		IntervalEncNode root = preorder.fetchNode();
		
		
		// get the next node
		IntervalEncNode next;
		if (preorder.next()) {
			next = preorder.fetchNode(); 
		} else {
			next = null;
		}
		
		// new values for anc and sib
		anc.removeLast(); 
		anc.addFirst(pqg.getHf().h(root.getLabel(), root.getValue()));
		LinkedList sib = pqg.getHf().getEmptyRegister(q);
		
		// if you are a leaf, produce a leafs pq-grams 
		if (root.isLeaf()) {
			pqg.addPQGram(root.getTreeID(), anc, sib);
		} else {
			// while the next nodes are your descendants 
			// tell them to produce their pq-grams
			while (root.isAncestor(next)) {
				sib.removeLast(); 
				sib.addFirst(pqg.getHf().h(next.getLabel(), next.getValue()));
				pqg.addPQGram(root.getTreeID(), anc, sib);
				next = getOrderedPQGrams(preorder, pqg, anc, q);
			}
			for (int i = 0; i < q - 1; i++) {
				sib.removeLast();
				sib.addFirst(pqg.getHf().getNullNode());
				pqg.addPQGram(root.getTreeID(), anc, sib);
			}
		}
		
		// return next
		return next;	
	}
	
	/**
	 * Calculates all pq-grams of the subtree rooted in the node n at
	 * the current position in "preorder" It has some side-effects:
	 *
	 * <ul>
	 * 
	 * <li>cursor in preorder is set to the nesuper.toString();xt node that is not a
	 * descendant of n (or to the root of the next tree, if n is the
	 * root node of a tree)
	 *
	 * </ul>
	 *
	 * @param anc next p ancestors of n
	 * @param sib next q left siblings of n
	 * @return next node in preorder
	 *
	 */
	private static IntervalEncNode getUnorderedPQGrams(Cursor preorder, PQGramTbl pqg, LinkedList<HashValue> stem_in, int q, int w) 
		throws SQLException {
		

		// copy input shift register (so it will not be changed)
		LinkedList<HashValue> stem = (LinkedList)stem_in.clone();
		
		// get current node
		IntervalEncNode root = preorder.fetchNode();		
		
		// get the next node
		IntervalEncNode next;
		if (preorder.next()) {
			next = preorder.fetchNode(); 
		} else {
			next = null;
		}
		
		// new values for anc and sib
		stem.removeLast(); 
		stem.addFirst(pqg.getHf().h(root.getLabel(), root.getValue()));
		
		// if you are a leaf, produce a leafs pq-grams 
		if (root.isLeaf()) {
			LinkedList<HashValue> base;
			base = pqg.getHf().getEmptyRegister(q);
			pqg.addPQGram(root.getTreeID(), stem, base);					
		} else {
			LinkedList<HashValue> children = new LinkedList<HashValue>();
			
			// while the next nodes are your descendants
			// - store the children
			// - tell them to produce their pq-grams
			while (root.isAncestor(next)) {
				children.add(pqg.getHf().h(next.getLabel(), next.getValue()));
				next = getUnorderedPQGrams(preorder, pqg, stem, q, w);
			}
			
			// copy children to array, add null nodes and sort
			int f = children.size();
			int dummies = Math.max(0, w - f);
			HashValue[] chArr = new HashValue[f + dummies];
			children.toArray(chArr);
			for (int i = 0; i < dummies; i++) {
				chArr[i + f] = HashValue.maxValue(pqg.getHf().getLength());
			}
			Arrays.sort(chArr);
			for (int i = 0; i < dummies; i++) {
				chArr[i + f] = pqg.getHf().getNullNode();
			}

			
			
			// produce leaf-parts of pq-grams
			for (int i = 0; i < chArr.length; i++) {
				LinkedList<HashValue> base = new LinkedList<HashValue>();
				if (q == 2) {
					base.add(chArr[i]);	  // starter node (first node in q-window)
					for (int j = i + 1; j < i + w; j++) {
						base.addFirst(chArr[j % chArr.length]);            // node in q-window other then starter node 
						pqg.addPQGram(root.getTreeID(), stem, base);
						base.removeFirst();
					}
				} else if (q == w) {
					for (int j = i; j < i + q; j++) {						
						base.addFirst(chArr[j % chArr.length]);  // node in q-window other then starter node
					}					
					pqg.addPQGram(root.getTreeID(), stem, base);					
				} else {
					throw new RuntimeException("Implementation incomplete. Works only for (q = 2, w >= 2) or (q > 2, w = q) -- SORRY.");
				}
			}
		}
		
		// return next
		return next;	
	}
	
	
	/**
	 * Calculates all pq-grams of the subtree rooted in the node n at
	 * the current position in "preorder" It has some side-effects:
	 *
	 * <ul>
	 * 
	 * <li>cursor in preorder is set to the next node that is not a
	 * descendant of n (or to the root of the next tree, if n is the
	 * root node of a tree)
	 *
	 * </ul>
	 *
	 * @param anc next p ancestors of n
	 * @param sib next q left siblings of n
	 * @return next node in preorder
	 *
	 */
	private static IntervalEncNode getUnorderedPQGrams(Cursor preorder, PQGramTbl pqg, LinkedList<HashValue> stem_in, int q, boolean split) throws SQLException {
		

		// copy input shift register (so it will not be changed)
		LinkedList<HashValue> stem = (LinkedList)stem_in.clone();
		
		// get current node
		IntervalEncNode root = preorder.fetchNode();		
		
		// get the next node
		IntervalEncNode next;
		if (preorder.next()) {
			next = preorder.fetchNode(); 
		} else {
			next = null;
		}
		
		// new values for anc and sib
		stem.removeLast(); 
		stem.addFirst(pqg.getHf().h(root.getLabel(), root.getValue()));
		
		// if you are a leaf, produce a leafs pq-grams 
		if (root.isLeaf()) {
			LinkedList sib;
			if (split) {
				sib = pqg.getHf().getEmptyRegister(2);
			} else {
				sib = pqg.getHf().getEmptyRegister(q);
			}
			pqg.addPQGram(root.getTreeID(), stem, sib);					
		} else {
			LinkedList children = new LinkedList();
			
			// while the next nodes are your descendants 
			// tell them to produce their pq-grams
			while (root.isAncestor(next)) {
				children.add(pqg.getHf().h(next.getLabel(), next.getValue()));
				next = getUnorderedPQGrams(preorder, pqg, stem, q, split);
			}
			
			// sort the children and add null nodes
			int f = children.size();
			int nulls = Math.max(0, Math.min(q - 1, 2 * q - f - 1));
			HashValue[] chArr = new HashValue[f + nulls];
			children.toArray(chArr);
			for (int i = f; i < f + nulls; i++) {
				chArr[i] = HashValue.maxValue(pqg.getHf().getLength());
			}
			//Arrays.sort(chArr);
			
			// produce leaf-parts of pq-grams
			for (int i = 0; i < f; i++) {
				if (split) {
					HashValue[] sibArr = new HashValue[2];
					sibArr[0] = chArr[i];	  // starter node (first node in q-window)
					for (int j = i + 1; j < i + q; j++) {
						sibArr[1] = chArr[j % chArr.length];  // node in q-window other then starter node 
						pqg.addPQGram(root.getTreeID(), stem, sortSib(sibArr, pqg.getHf()));
					}
				} else {
					HashValue[] sibArr = new HashValue[q];
					for (int j = i; j < i + q; j++) {						
						sibArr[j - i] = chArr[j % chArr.length];  // node in q-window other then starter node
					}					
					pqg.addPQGram(root.getTreeID(), stem, sortSib(sibArr, pqg.getHf()));					
				}
			}
		}
		
		// return next
		return next;	
	}
	
	private static LinkedList sortSib(HashValue[] sibArr, FixedLengthHash hf) {
		Arrays.sort(sibArr);
		LinkedList sib = new LinkedList();
		for (int i = 0; i < sibArr.length; i++) {
			if (sibArr[i].equals(HashValue.maxValue(hf.getLength()))) {
				sib.addFirst(hf.getNullNode());
			} else {
				sib.addFirst(sibArr[i]);
			}
		}
		return sib;
	}
	
	
 	public static void main(String[] args) throws Exception {
 		if (args.length < 4) {
 			System.out.println("params: p q w xml-file(s)");
 			System.exit(-1);
 		}
 		
 		// establish DB connection
		MySqlTools.loadDriver();
		Connection out = MySqlTools.getConnection("localhost", "test",
			 	"user", "password");
		Connection in = MySqlTools.getConnection("localhost", "test",
			 	"user", "password");

		// get parameters
		int p = Integer.parseInt(args[0]);
		int q = Integer.parseInt(args[1]);
		int w = Integer.parseInt(args[2]);

		// load xml files into database
		long time;
		IntervalEncForest f = new IntervalEncForest(out, in, 
				new MySqlInsertBuffer(), "forest");	
		f.open();
		f.reset();
		System.out.println("Loading XML file(s) into DB (interval encoding)...");
		for (int i = 3; i < args.length; i++) {
			time = System.currentTimeMillis();
			System.out.print("  " + args[i] + "...");
			xml.IntervalEncHandler.parseFromFile(args[i], f, i - 1, 
					true, true);
			System.out.println((System.currentTimeMillis() - time) + "ms");
		}		
		// creating indexes on forest attributes
		time = System.currentTimeMillis();
		System.out.print("  Creating index on forest...");
		f.createPreorderIndex();
		System.out.println((System.currentTimeMillis() - time) + "ms");
		f.close();
		
		// create empty profile size table
		ProfileSizeTbl ps = new ProfileSizeTbl(out, "tblProfSize");
		
		// compute the unordered pq-grams
		System.out.print("Computing unorderd sorted pq-grams...");
		PQGramTbl pqgu = new PQGramTbl(out, new MySqlInsertBuffer(), "tblPQGramsUnord", 
				new PrefixHash(6), p, 2);
		pqgu.reset();	
		ps.reset();
		time = System.currentTimeMillis();
		getUnorderedPQGrams(f.forestInPreorder(), pqgu, ps, p, q, w);
		System.out.println((System.currentTimeMillis() - time) + "ms");
		pqgu.close();

		

		// compute the ordered pq-grams		
		System.out.print("Computing orderd pq-grams...");
		PQGramTbl pqgo = new PQGramTbl(out, new MySqlInsertBuffer(), "tblPQGramsOrd", 
				new PrefixHash(6), p, q);
		pqgo.reset();
		ps.reset();
		time = System.currentTimeMillis();
		getPQGrams(f.forestInPreorder(), pqgo, ps, p, q);
		System.out.println((System.currentTimeMillis() - time) + "ms");
		pqgo.close();
		f.dropPreorderIndex();
	}

}
