/**
 * 
 */
package tasmTED;

import distance.SubtreeWeighting;
import distance.WeightedEditDist;
import distance.WeightingFunction;
import tree.LblValTree;
import util.Heap;

/**
 * @author Nikolaus Augsten
 *
 */
public class TASMPostorder extends TASM implements PostorderQueue {

	public static boolean DEBUG = false;
	
	// initilization
	private WeightingFunction weightingFunction1, weightingFunction2;
	private PostorderSource postorderSource;
	
	// input parameters to TASM
	private TEDTree que;
	private RingBuffTEDTreeKR streamDoc;
	private int k;
	
	// variables used by TASM computation
	private Heap topK;
	private int computeNext = 0;
	
	/**
	 * @param wQuery
	 * @param wDoc
	 * @param postorderSource
	 */
	public TASMPostorder(WeightingFunction wQuery, WeightingFunction wDoc, 
			PostorderSource postorderSource) {
		this.weightingFunction1 = wQuery;
		this.weightingFunction2 = wDoc;
		this.postorderSource = postorderSource;
	}

	/* (non-Javadoc)
	 * @see tasmTED.TASM#tasm(distance.TEDTree, distance.TEDTree, int)
	 */
	@Override
	public Heap tasm(TEDTree que, TEDTree doc, int k) {
		if (k == 0) return new Heap(0);
		
		// copy parameters
		int maxSubtreeSize = (int)Math.round(Math.ceil(que.getNodeCount()  
				+ weightingFunction1.maxNodeWeightSum(que.getNodeCount()) 
				+ weightingFunction2.maxNodeWeightSum(k)));
		this.que = que;
		this.k = k;
		this.streamDoc = new RingBuffTEDTreeKR(maxSubtreeSize);
		
		// initialize
		this.topK = new Heap(k);
		this.computeNext = 1;
		
		// start computation
		postorderSource.appendTo(this);
		
		// compute topK for root node and merge with previous topK list
		this.topK.merge( 
			subtreeTASM(this.que, this.streamDoc, this.k, 
					this.computeNext, this.streamDoc.root(),
					new TASMDynamic(this.weightingFunction1, 
							this.weightingFunction2)));

		
		return topK;
	}


	private static Heap subtreeTASM(TEDTree que, RingBuffTEDTreeKR doc, 
			int k, int computeNext, int subtreeRoot, TASMDynamic tasmDynamic) {
		
		Heap topK = new Heap(k);
		int U = doc.getBufferSize() - que.getNodeCount();
		
		while (subtreeRoot >= computeNext) {
			int subtreeSize = subtreeRoot - doc.lld(subtreeRoot) + 1;		
			int L = subtreeSize - que.getNodeCount();
			if (L <= U && (topK.getSize() < k || 
					L < ((NodeDistPair)topK.peek()).getDist())) {
				// get subtree
				RingBuffTEDTree subtree = 
					doc.getSubtree(subtreeRoot);
				if (DEBUG) System.out.println("subtree starting at " + computeNext + ": " + subtree);				
				// merge topK for subtree with previous topK list
				Heap subtreeTopK = tasmDynamic.tasm(que, subtree, k);
				if (DEBUG) System.out.println("topK of subtree: " + subtreeTopK);
				topK.merge(tasmDynamic.tasm(que, subtree, k));
				subtreeRoot -= subtree.getNodeCount();
			} else {
				subtreeRoot--;
			}
			
		}
		if (DEBUG) System.out.println("topK of subtreeTASM call: " + topK);
		return topK;
	}
	
	/* (non-Javadoc)
	 * @see tasmTED.PostorderQueue#append(java.lang.String, int)
	 */
	public void append(String label, int subtreeSize) {
		if (DEBUG) System.out.println((streamDoc.getNodeCount() + 1) + ": (" + label + "," + subtreeSize + ")");
		if (streamDoc.getNodeCount() >= streamDoc.getBufferSize() && // ringbuffer is already filled
				computeNext == streamDoc.smallestValidID()) { // next node would shift out uncomputed node

			// next uncomputed node is a leaf
			if (streamDoc.lld(computeNext) == computeNext) { 

				// get root of leafmost subtree
				int rootNext = streamDoc.getLeafKeyRoot(computeNext);
				// compute TASM for leafmost subtree
				Heap subtreeTopK = subtreeTASM(this.que, this.streamDoc, 
						this.k, this.computeNext, rootNext,
						new TASMDynamic(this.weightingFunction1, 
								this.weightingFunction2));

				// merge topK of subtrees with current topK
				topK.merge(subtreeTopK);
				if (DEBUG) System.out.println("global topK after subtreeTASM call: " + this.topK);

				// set "this.computeNext" to node after leftmost subtree root
				computeNext += rootNext - computeNext + 1;	
			} else { // next uncomputed node is the root of a subtree that can be pruned
				if (DEBUG) System.out.println("skipping computeNext (non-leaf): " + computeNext + " " + streamDoc.getLabel(computeNext));
				computeNext++; 
			}
			
		}
		// append this node
		streamDoc.append(label, subtreeSize);
		
	}

	public static void main(String[] args) {
		int k = Integer.parseInt(args[0]);
		LblValTree query = LblValTree.fromString("1:{a{b}{d}}");
		LblValTree doc = LblValTree.fromString("2:{root{a{b}{c}}{a{c}{b}}{x{b}{c}}{a{b{x}{y}}{d}}{x{a{b}{d}}}{a{x}{d}}}");
		WeightedEditDist dist = new WeightedEditDist(
				new SubtreeWeighting(0), new SubtreeWeighting(0), false);
		TASMPostorder tasmPostorder = 
			new TASMPostorder(dist.getWeightingFunction1(), 
					dist.getWeightingFunction2(), 
					new StaticTEDTree(doc));
		Heap topK = tasmPostorder.tasm(new StaticTEDTree(query), null, k);
		System.out.println("topK: " + topK);
	}
	
}
