/**
 * 
 */
package tasmalgos;

import java.sql.SQLException;
import java.util.Enumeration;

import distance.SubtreeDist;
import distance.TreeDist;
import distance.WeightedEditDist;
import intervalenc.Cursor;
import intervalenc.IntervalEncNode;
import tasmTED.NodeDistPair;
import tasmTED.StaticTEDTree;
import tree.LblValTree;
import util.Heap;

/**
 * <p>Collection of static methods with the following signature:</p>
 * <p><code>tasmNN(query, document, k, distanceFunction)</code></p>
 * <p>Computes the top-<code>k</code> matches of the <code>query</code> tree
 * to a subtree of the <code>document</code> tree using
 * <code>distanceFunction</code> for the ranking. The trees are ordered and labeled.
 * @author naugsten
 */
public class TASM {

	public static boolean DEBUG = false;
	
	/**
	 * The cursor points to the root node of the subtree. After the execution, it points to the 
	 * node in preorder after the root node.
	 * 
	 * @param cursor
	 * @return
	 * @throws SQLException
	 */
	public static LblValTree getSubtree(Cursor cursor) throws SQLException {
		IntervalEncNode root = cursor.fetchNode();
		LblValTree n = new LblValTree(root.getLabel(), root.getValue(), root.getTreeID());
		n.setNodeID(root.getLft());
		cursor.next();
		while (!cursor.isAfterLast() && root.isAncestor(cursor.fetchNode())) {
			n.add(getSubtree(cursor));
		}
		return n;
	}
	
	/**
	 * Exectutes the distance function only between the query and the document tree. 
	 * The distances to the subtrees are computed as a side product.
	 * Both query and document are materialized as a {@link tree.LblValTree} in main memory.
	 * 
	 * @param query
	 * @param document
	 * @param k
	 * @param dist
	 * @return
	 * @throws SQLException
	 */
	public static NodeDistPair[] tasmSubtreeDist(Cursor query, Cursor document, int k, SubtreeDist dist) throws SQLException {
		if (k == 0) {
			return new NodeDistPair[0];
		}
		// Load trees
		query.next();
		LblValTree q = TASM.getSubtree(query);
		document.next();
		LblValTree doc = TASM.getSubtree(document);
		// compute all distances
		double[] dists = dist.subtreeDists(new StaticTEDTree(q), new StaticTEDTree(doc));
		// select top-k
		Heap heap = new Heap(k);
		int i = 1;
		for (Enumeration<LblValTree> e = doc.postorderEnumeration(); e.hasMoreElements();) {
			int nodeID = e.nextElement().getNodeID();					
			if (heap.getSize() < k || dists[i] < ((NodeDistPair)heap.peek()).getDist()) {
				NodeDistPair result = new NodeDistPair(nodeID, dists[i]);
				if (!heap.insert(result)) {
					heap.substitute(result);
				}
			}
			i++;
		}
		return heapToArray(heap);
	}
	
	/**
	 * Executes the distance function between the query tree and each subtree of the document.
	 * Both query and document are materialized as a {@link tree.LblValTree} in main memory.
	 * 
	 * @param query
	 * @param document
	 * @param k
	 * @param dist
	 * @return
	 * @throws SQLException
	 */
	public static NodeDistPair[] tasmBruteForce(Cursor query, Cursor document, int k, TreeDist dist) throws SQLException {
		if (k == 0) {
			return new NodeDistPair[0];
		}
		// Load trees
		query.next();
		LblValTree q = TASM.getSubtree(query);
		document.next();
		LblValTree doc = TASM.getSubtree(document);
		// compute distances, max-heap maintains top-k
		Heap heap = new Heap(k);
		for (Enumeration<LblValTree> e = doc.depthFirstEnumeration(); e.hasMoreElements(); ) {
			LblValTree subtree = e.nextElement();
			double d = dist.treeDist(q, subtree);
			if (heap.getSize() < k || d < ((NodeDistPair)heap.peek()).getDist()) {
				NodeDistPair result = new NodeDistPair(subtree.getNodeID(), d);
				if (!heap.insert(result)) {
					heap.substitute(result);
				}
			}
		}
		return heapToArray(heap);
	}
		
	private static void debugPrint(IntervalEncNode currentNode, double lowerBound, double currentUpperBound) {
		System.out.print("(" + currentNode.getLabel() + "," + currentNode.getLft() + "," + currentNode.getRgt() + ")");
		int size = (currentNode.getRgt() - currentNode.getLft() + 1) / 2 ;
		System.out.print(" |t(v)|=" + size);
		System.out.print(" L=" + lowerBound + " U=" + currentUpperBound);
	}
 	
	public static NodeDistPair[] tasm(Cursor query, Cursor document, 
			int k, WeightedEditDist dist) throws SQLException {
		if (k == 0) {
			return new NodeDistPair[0];
		}
		
		// make a tree of the query cursor
		query.next();
		LblValTree q = TASM.getSubtree(query);
		int querySize = q.getNodeCount();
		double globalUpperBound = dist.maxDist(querySize, k);
		Heap heap = new Heap(k);
		document.next();		
		// traverse the whole document in preorder
		while (!document.isAfterLast()) {
			IntervalEncNode currentNode = document.fetchNode(); // current subtree
			int lowerBound = Math.abs(currentNode.subtreeSize() - querySize); 
			// Is the lower bound for the distance between the query and the current subtree  
			//   smaller than the maximum distance that can produce a match?
			if (DEBUG) {
				debugPrint(currentNode, lowerBound, 
						heap.isEmpty() ? globalUpperBound : ((NodeDistPair)heap.peek()).getDist());
			}
			if (lowerBound <= globalUpperBound &&
					(heap.getSize() < k || lowerBound < ((NodeDistPair)heap.peek()).getDist())) {
				LblValTree subDoc = getSubtree(document);
				// Compute the distance between the query and all subtrees of current subtree.
				// dists[0] is the distance to the first node of the current subtree in POSTorder.
				double[] dists = dist.subtreeDists(new StaticTEDTree(q), new StaticTEDTree(subDoc));
				if (DEBUG) {
					System.out.println(" ***** dist=" + dists[subDoc.getNodeCount() - 1] + subDoc);
				}
				// update top-k on heap
				int i = 1;
				for (Enumeration<LblValTree> e = subDoc.postorderEnumeration(); e.hasMoreElements();) {
					int nodeID = e.nextElement().getNodeID();					
					if (heap.getSize() < k || dists[i] < ((NodeDistPair)heap.peek()).getDist()) {
						NodeDistPair result = new NodeDistPair(nodeID, dists[i]);
						if (!heap.insert(result)) {
							heap.substitute(result);
						}
					}
					i++;
				}
			} else {
				document.next();
				if (DEBUG) {
					System.out.println();
				}
			}
		}		
		return heapToArray(heap);
	}
	
	private static NodeDistPair[] heapToArray(Heap heap) {
		Comparable[] copyOfHeap = heap.getSortedArray();
		NodeDistPair[] topK = new NodeDistPair[copyOfHeap.length];
		for (int i = 0; i < copyOfHeap.length; i++) {
			topK[i] = (NodeDistPair)copyOfHeap[i];
		}
		return topK;
	}
	
}
