package tree;

import hash.FixedLengthHash;
import hash.HashValue;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeNode;

import adjacencyenc.AdjacencyEncForest;

import util.FormatUtilities;

import java.util.Enumeration;
import java.util.Vector;


/**
 * A node of a tree. Each tree has an ID.
 * The label can be empty, but can not contain trailing spaces (nor consist only of spaces).
 * 
 * Two nodes are equal, if there labels are equal, and n1 < n2 if label(n1) < label(n2).
 */
public class LblTree extends DefaultMutableTreeNode implements Comparable {
	
	public static final String TAB_STRING = "    ";
	public static final String ROOT_STRING =   "*---+";
	public static final String BRANCH_STRING = "+---+";
	
	public static final String OPEN_BRACKET = "{";
	public static final String CLOSE_BRACKET = "}";
	public static final String ID_SEPARATOR = ":";
	
	public static final int HIDE_NOTHING = 0;
	public static final int HIDE_ROOT_LABEL = 1;
	public static final int RENAME_LABELS_TO_LEVEL = 2;
	public static final int HIDE_ALL_LABELS = 3;
	public static final int RANDOM_ROOT_LABEL = 4;
	
	int treeID = Node.NO_TREE_ID;
	String label = null;
	Object tmpData = null;
	int nodeID = Node.NO_NODE;
	
	/**
	 * Use only this constructor!
	 */
	public LblTree(String label, int treeID) {
		super();
		this.treeID = treeID;
		this.label = label;
	}
		
	public void setLabel(String label) {
		this.label = label;
	}
	
	public String getLabel() {
		return label;
	}
	
	public String showNode() {
		return label;
	}
	
	public int getTreeID() {
		if (isRoot()) {
			return treeID;
		} else {
			return ((LblTree)getRoot()).getTreeID();
		}
	}
	
	public void setTreeID(int treeID) {
		if (isRoot()) {
			this.treeID = treeID;
		} else {
			((LblTree)getRoot()).setTreeID(treeID);
		}
	}
	
	/**
	 * tmpData: Hook for any data that a method must attach to a tree.
	 * Methods can assume, that this date is null and should return it
	 * to be null!
	 */
	public void setTmpData(Object tmpData) {
		this.tmpData = tmpData;
	}
	
	public Object getTmpData() {
		return tmpData;
	}
	
	/**
	 * Clear tmpData in subtree rooted in this node.
	 */
	public void clearTmpData() {
		for (Enumeration e = breadthFirstEnumeration(); e.hasMoreElements();) {
			((LblTree)e.nextElement()).setTmpData(null);
		}
	}
	
	
	public String toLatex() {
		String res = "";
		String label = util.FormatUtilities.escapeLatex(this.showNode());
		if (isLeaf() && !isRoot()) {
			res += "\\Tr{" + label + "}";
		} else {
			int levelsep = 20 + (int)(0.3 * this.getNodeCount());
			res += "\\pstree[linewidth=0.2pt,levelsep=" + levelsep + "pt,treefit=tight,treesep=4pt,nodesep=2pt]{\\Tr{" + label + "}}{";
			for (Enumeration e = children(); e.hasMoreElements();) {
				res += ((LblTree)e.nextElement()).toLatex() + "\n";
			}	   
			res += "}";
		}
		return res;
	}
	
	public void prettyPrint() {
		prettyPrint(false);
	}
	
	public void prettyPrint(boolean printTmpData) {
		for (int i = 0; i < getLevel(); i++) {
			System.out.print(TAB_STRING);
		}
		if (!isRoot()) {
			System.out.print(BRANCH_STRING);
		} else {
			if (getTreeID() != Node.NO_TREE_ID) {
				System.out.println("treeID: " + getTreeID());
			}
			System.out.print(ROOT_STRING);
		}
		System.out.print(" '" + this.showNode() + "' ");
		if (printTmpData) {
			System.out.println(getTmpData());
		} else {
			System.out.println();
		}
		for (Enumeration e = children(); e.hasMoreElements();) {
			((LblTree)e.nextElement()).prettyPrint(printTmpData);
		}
		
	}
	
	public int getNodeCount() {
		int sum = 1;
		for (Enumeration e = children(); e.hasMoreElements();) {
			sum += ((LblTree)e.nextElement()).getNodeCount();
		}
		return sum;
	}
	
	public double getMediumFanout() {
		int nodes = getNodeCount();
		int parents = nodes - getLeafCount();
		int children = nodes - 1;
		if (parents == 0) {
			return 0.0;
		} else {
			return ((double)children / (double)parents);
		}
	}
	
	/**
	 * @return all leafs from left to right.
	 */
	public LblTree[] getLeafs() {
		int leafCount = getLeafCount();
		LblTree[] leafs = new LblTree[leafCount];
		int i = 0;
		for (Enumeration e = depthFirstEnumeration(); e.hasMoreElements();) {
			LblTree n = (LblTree)e.nextElement();
			if (n.isLeaf()) {
				leafs[i] = n;
				i++;
			}
		}
		return leafs;
	}
	
	/**
	 * @param enums enumeration to use (0=breadthFirst,
	 * 1=depthFirst=postorder, 2=preOrder)
	 * @return array with all internal nodes
	 */
	public LblTree[] getInternalNodes(int enums) {
		int leafCount = getLeafCount();
		int nodes = getNodeCount();
		LblTree[] internal = new LblTree[nodes - leafCount];
		int i = 0;
		Enumeration e;
		switch (enums) {
		case 1: e = depthFirstEnumeration(); break; 
		case 2: e = preorderEnumeration(); break; 
		case 0: 
		default:
			e = breadthFirstEnumeration(); break;
		}	
		while (e.hasMoreElements()) {
			LblTree n = (LblTree)e.nextElement();
			if (!n.isLeaf()) {
				internal[i] = n;
				i++;
			}
		}
		return internal;
	}
	
	/**
	 * Constructs an LblTree from a string representation of tree. The
	 * treeID in the String representation is optional; if no treeID is given,
	 * the treeID of the returned tree will be NO_ID.
	 *
	 * @param s string representation of a tree. Format: "treeID:{root{...}}".
	 * @return tree represented by s
	 */
	public static LblTree fromString(String s) {
		int treeID = FormatUtilities.getTreeID(s);
		//System.out.println(s);
		s = s.substring(s.indexOf(OPEN_BRACKET), s.lastIndexOf(CLOSE_BRACKET) + 1);
		LblTree node = new LblTree(FormatUtilities.getRoot(s), treeID);
		Vector c = FormatUtilities.getChildren(s);
		//System.out.println("vector "+c.size());
		//System.out.println("S "+s.length()+" "+s);
		for (int i = 0; i < c.size(); i++) {
			node.add(fromString((String)c.elementAt(i)));
		}
		return node;
	}
	
	/**
	 * String representation of a tree. Reverse operation of {@link #fromString(String)}.
	 * treeID is NO_ID, it is skiped in the string representation.
	 *
	 * @return string representation of this tree
	 *
	 */
	@Override
	public String toString() {	
		String res = OPEN_BRACKET + showNode();
		if ((getTreeID() >= 0) && (isRoot())) {
			res = getTreeID() + ID_SEPARATOR + res;
		}
		for (Enumeration e = children(); e.hasMoreElements();) {
			res += ((LblTree)e.nextElement()).toString();
		}
		res += CLOSE_BRACKET;
		return res;
	}   
	
	
	/**
	 * 
	 */
	public static LblTree deepCopy(LblTree t) {
		LblTree nt = new LblTree(t.getLabel(), t.getTreeID());
		nt.setTmpData(t.getTmpData());
		for (Enumeration e = t.children(); e.hasMoreElements(); ) {
			nt.add(deepCopy((LblTree)e.nextElement()));
		}
		return nt;
	}
	

	/**
	 * Returns the number of the node in the sibling set.
	 * 
	 * {@link DefaultMutableTreeNode.getIndex(DefaultMutableTreeNode)} fails
	 * if nodes have the same label, as it uses the {@link #equals(Object)} method
	 * to find a node. This method compares the object addresses.
	 *
	 */
	@Override
	public int getIndex(TreeNode n) {
		for (int i = 0; i < this.getChildCount(); i++) {
			if (this.getChildAt(i) == n) {
				return i;
			}
		}
		return -1;
	}
	
	/**
	 * Compares the labels.
	 */
	public int compareTo(Object o) {		
		return getLabel().compareTo(((LblTree)o).getLabel());
	}
	

	/**
	 * Compares the labels.
	 */
	@Override
	public boolean equals(Object o) {
		return this.compareTo(o) == 0;
	}

	/* (non-Javadoc)
	 * @see tree.EditTree#deleteNode(java.lang.Object)
	 */
	public boolean deleteNode(Object nodeOfTree) {
		try {
			LblTree n = (LblTree)nodeOfTree;
			if (n.isRoot()) {
				return false;				
			} else {
			    LblTree parent = (LblTree)n.getParent();
			    int nChildID = parent.getIndex(n);
			    n.removeFromParent();
			    int i = 0;
			    LblTree[] childArr = new LblTree[n.getChildCount()];
			    for (Enumeration<LblTree> e = n.children(); e.hasMoreElements();) {
			    	childArr[i] = e.nextElement();
			    	i++;
			    }
			    for (i = 0; i < childArr.length; i++) {
			    	parent.insert(childArr[i], nChildID + i); 
			    }
			    return true;
			}			
		} catch (ClassCastException e) {
			return false;
		}
	}

	/* (non-Javadoc)
	 * @see tree.EditTree#insertNode(java.lang.Object, java.lang.Object, int, int)
	 */
	public boolean insertNode(Object newNode, int k, int n) {
		LblTree nd = (LblTree)newNode;
		if ((this.getChildCount() >= k + n - 1) &&      	// do all children v_k ... v_(k+n-1) exist?
				(nd.getRoot() != this.getRoot())) {      // n and p can not be of the same tree
			for (int j = 0; j < n; j++) {
				nd.add((LblTree)this.getChildAt(k - 1));
			}
			this.insert(nd, k - 1);
			return true;
		} else {
			return false;
		}
	}
	
	/* (non-Javadoc)
	 * @see tree.EditTree#insertPath(java.lang.Object, java.lang.Object)
	 */
	public boolean insertPath(Object path) {
		LblTree p = (LblTree)path;
		int cmp = this.compareTo(p);

		if (cmp == 0) {    // follow down the path
			if (!p.isLeaf()) {
				int i = 0;
				// search a match among children
				// linear search - slow!!!
				while ((i < this.getChildCount()) &&  
						(!((LblTree)this.getChildAt(i)).insertPath(p.getFirstChild()))) {
					i++;
				}
				if (i == this.getChildCount()) {
					this.add((LblTree)p.getFirstChild());
				}
			}
			return true;
		} else if (cmp > 0) {
			if (!isRoot()) {
				LblTree parent = ((LblTree)this.getParent()); 
				int i = parent.getIndex(this);     // slow, as getIndex() performs linear search in children array
				parent.insert(p, i);
				return true;
			} else {
				return false;
			}
		}
		return false; 		// cmp < 0
	}

	/* (non-Javadoc)
	 * @see tree.EditTree#deletePath(java.lang.Object, java.lang.Object)
	 * 
	 * Problem: returns true, if tree should be empty!
	 * 
	 */
	public int deletePath(Object path) {
		LblTree n = (LblTree)path;
		int cnt = 0;
		while (n.isLeaf() && !n.isRoot()) {
			LblTree l = n;
			n = (LblTree)n.getParent();
			l.removeFromParent();
			cnt++;
		}
		if (n.isLeaf() && n.isRoot()) {
			return -1;
		} else {
			return cnt;
		}
	}
	
	
	/* (non-Javadoc)
	 * @see tree.EditTree#updateNode(java.lang.Object, java.lang.String)
	 */
	public boolean relabelNode(String label) {
		if (!getLabel().equals(label)) {
			setLabel(label);	
			return true;
		} else {
			return false;
		}
	}
	
	public HashValue getNodeHash(FixedLengthHash hf) {
		return hf.getHashValue(this.getLabel());
	}
	
	/**
	 * Taylormade to addresses: 
	 * * all empyt leafs at the 1st (house number) and 3th level (apartment numbers) are removed
	 * * on the other levels, empty leaves are removed if they are the only child of their parent
	 */
	public void cleanNullValues() {
		LblTree[] ch = new LblTree[this.getChildCount()];
		int i = 0;
		for (Enumeration e = this.children(); e.hasMoreElements();) {
			ch[i] = (LblTree)e.nextElement();
			i++;
		}
		for (i = 0; i < ch.length; i++) {
			ch[i].cleanNullValues();
		}
		if (getLabel().equals("") && (isLeaf())) {
				if (getLevel() != 2) {
					removeFromParent();
				} else if (getParent().getChildCount() == 1) {
					removeFromParent();
				} 
		}
	}
	
	public static void main(String[] args) {
		LblTree t = LblTree.fromString("1:{mg{rq{gw{rq{md}{sn}{lr}}{sn}{ob}{lr{bz}{ob}{sn}{rq}{ob{bz}{uk}}{bz}}{bz}{lr}{rq}}{uk}{sn}{md{md{rq}{md}{lr}{cm}{md}{sn}{mg}{mg}{ob}}{mg{lr}{uk}{sn}{mg}{cm{lr}{gw}{uk}{uk}{gw}}{uk}{uk{md}{rq}{ob}{lr{cm}{gw}{uk}{sn{lr}}{lr{mg}{sn}{mg}{ob}{lr}{sn}{bz}{mg}{gw}{gw}}{sn}{lr}}}{mg}}{md}}}{rq{cm}{rq}}{rq{uk}{ob{ob}}{lr}{mg}{lr}{bz}}{md{gw}{uk{sn{rq{gw}{gw}{lr}{md{cm}}{uk{md}{uk{lr}{lr}{sn}{uk}{ob}{rq}{cm}{ob}{uk}{gw{uk}{ob}{uk}{cm}}}{md}{lr}{cm}}}{ob{bz}{uk{lr{uk{cm}{uk{sn{bz}}{mg{bz}{cm}{rq}{ob}}{cm}}}{sn{bz}{md{rq}{uk}{ob}}}{lr}{cm{bz{uk}{bz}{gw}{ob{cm}{rq{md}{uk}}}{lr{ob}{lr}}{md{md{bz}{md}{bz}{cm}}}{md{sn}}{ob{ob}{cm}{md{gw}{md}{uk}}{sn}{lr}{lr}{sn}{md{sn}{mg}{rq}{cm}}{sn}}{cm}}{ob{uk}{bz}}{sn{sn{ob}}{uk{bz}{sn}{gw}{uk}{sn}{md{lr}{bz}{mg}{ob{gw}{mg}}}{sn{bz}{ob}{lr{md}{rq}{gw}{uk}{cm}{lr}}{lr{cm}{mg}{lr}{sn}{sn}{bz}{cm}{mg}}}{lr}{rq{ob}{uk}{cm}{bz}{rq{lr}{bz}{uk}{rq}{ob}}{uk{mg}{sn}{rq}{sn{ob}{bz}{mg}{ob}{uk}{sn}}}{mg}}{md}}}{uk}{ob{uk}{sn}}}{cm{bz}{md}{ob{rq}{lr}{sn}{uk{uk}{rq}{md}{gw}{rq}{ob}{lr}}{sn}{ob{bz}{md}{mg}{sn{gw}{rq}{ob}{gw}}{uk}{uk}{lr}}{gw{sn}{lr}{rq}{md}}{cm}{mg}{lr}}}{ob{rq}{cm}}{cm{ob{ob}}}{ob{sn}{lr}{cm{ob}{sn}{lr}{uk}}{ob}{mg{bz}{gw}{mg}{lr}{md}{bz}}{uk{cm}{mg{lr}{md}{gw}{ob}{ob{uk}{cm}}{cm}{cm}{sn}{bz}}{md}{bz{rq}{cm}{bz}}{rq{lr}{sn}{mg}{bz{ob}}{sn{sn}}}{mg}{mg}{md}{rq}}{md}{sn{cm}}{sn{gw{uk}}{rq{gw}{mg}}}{gw{uk}{uk}{ob}}{cm}{bz{ob}{mg{lr}{lr}{gw}{sn{ob{rq}{mg}{sn}{bz}{rq{rq}{bz}}{gw}}{md}}{md}{uk}}{lr{md}}{cm{ob}{gw}{lr{sn}{cm}{cm}{cm}}{md}{rq{md}{uk}{mg}}{mg}{lr{mg}{gw{mg}{lr}{bz}{mg}}}}}}{ob{mg{cm}{lr}{cm{uk{gw}{lr}{uk}{mg}{md}}}}{ob{rq}{cm}{uk}{cm}{uk}{lr}}}{cm}}{uk{ob{uk}{uk{uk}{gw{uk}{cm}{uk}{lr}{mg}}{mg}}{md{mg}{mg}{md}{ob}{cm}{uk}{rq}}{rq{sn}{lr{sn{mg}{mg}{cm}{cm}{md}}{ob}{gw}{lr}{mg}}{mg}{uk}{rq{lr}{cm{sn}{uk}{ob}{mg}{bz}{bz}}{mg{rq}}{bz{ob}{uk}{mg}}{lr}{gw}{gw{lr}{md}}{gw}{md}{sn}}}{rq}}{gw{bz}{gw}{uk{sn}{mg}{bz}}{cm}}}}{rq{rq{ob}{rq}{md}{bz}{bz}{sn}{cm}{ob}{uk}{bz}{mg}{gw}}{gw}{ob}{cm}{mg}{uk{mg}{bz}{ob}{uk}{gw}{uk}{ob}}}{uk{gw{rq}{gw}{bz{mg}{sn}{gw}}}{md{mg}{sn{mg}{rq{rq}{rq}}{cm{ob}{gw}}{lr{sn}{lr}{md}{bz}{sn}{mg}{lr}{gw}}{cm{sn}}{cm{rq}{bz}{ob}}{rq}{cm{lr}{ob}{cm}}{md}}{bz}{sn{ob}{rq{gw}}{rq{mg{sn}{mg}{ob}{cm}{cm}{cm}{cm}{lr}}{sn{mg}}{uk{lr}{lr}{mg}{gw}{md{cm}{gw}{gw}}}{bz}{sn{ob}{mg}{cm{ob}{gw}{rq}}{rq}{lr}{md}{lr}{rq{md}{sn}{uk}}}{gw{bz}{ob}{mg}{ob}}{rq}{lr}}}{gw}{rq}{lr}}}{rq{sn{lr}{lr}{bz}}{lr{md{lr}{gw}}{bz}}{ob}{sn{lr{sn}{sn{mg{cm}{lr}{mg}{bz}{mg}{uk}{mg}{mg{sn}}}{bz}{md}{uk}{md}{rq}}{mg}{lr}{bz}{gw}{gw{cm}{sn{lr}}{uk{lr}{uk{md}{lr}{lr}}{ob{rq}{gw}{lr}{cm}{sn}}{bz}{cm{mg}{bz{md}{bz}{cm{gw}{mg{cm}{gw}}{uk}{md{bz}{gw}{gw}{ob}}{uk{lr}{gw}{lr}{uk}{mg}{uk}{rq}{bz}{mg}}{bz{rq}{md}{bz}{lr}}}{sn}}{sn}{gw{md}{md}{bz}{gw}{uk}{uk}}}}{ob}{md{lr}{sn}{bz}{gw{mg}{md}{sn}}{sn}{uk}{sn{lr}{cm{lr}{md}{md}{rq}{bz}{cm}{uk}}{sn}{md}{bz}{ob{gw}{mg}{cm}{uk}{lr}{uk}{mg}{lr}{bz{md}{lr}{rq}{uk}{gw}{gw}{ob}{sn}{bz}{uk}{gw}{bz}{rq}}{rq}}}{bz{cm}{gw}{ob}{ob}}{gw}{lr}}}{gw{lr}}{ob{rq{gw}{lr}{lr}{ob}{rq{rq{lr{mg}}{cm{mg}}{gw{md}{md}{mg}{mg}{uk}{rq{mg}}{lr{gw}{md}{rq}{uk}{bz}}{uk}}{gw}{sn}{rq}{mg}{bz}{uk}{ob}{lr}{sn}{mg{uk}{mg}{md}{rq}{uk}{bz{gw}}{cm{bz}{bz}{uk}{cm}{md}}{sn}{ob}{mg}{mg}{mg}}{lr}{ob{sn}}{md{ob}{md}{lr}{uk{mg}{bz}{sn}{mg}{rq}{cm}}{lr}{bz}{rq}}}{ob}{ob}{bz}{md}{gw}}{lr}{mg{cm}}{cm}{cm}}{md{rq{uk}{sn}}{ob{mg}{rq{sn}{rq}{sn}{cm}{rq}{cm}{gw{mg}}{md{gw}}{lr{cm}{sn}{uk}}{ob}}{cm}}{bz}{cm}{rq}{ob}{bz}{bz}{lr}{bz}{rq}{ob}{bz}}{lr}{ob}{uk}{sn}{sn{md{mg}{bz}{sn}{uk}}{lr{ob}{gw}}{lr}{lr}{ob}{rq{sn}{sn{lr}}{lr{rq}{bz}{bz}{sn}{mg}}{gw}{sn}{cm}{bz}}}{mg}{bz}}{bz{cm{lr{lr}{lr}{mg}{ob}}{uk}}}}{sn{sn}{lr}{ob}{uk}{ob{sn{cm}}{lr{ob}{ob}}}{lr}}{uk}}{bz}{bz{uk{lr}{rq}{lr}{rq}}{md}}{uk{rq{mg}}{mg{sn}{sn}{uk}{ob}{md}{mg}{sn}{ob}{rq}{rq}{rq{sn}{gw}{sn}{sn{gw}}{lr{md}}{bz}{bz}}{cm}}{mg{uk}}{gw{mg}{md}{rq}{rq}{sn}}{mg}}}}{gw{rq}{cm{mg{mg}{md}{ob}}{rq}{bz}}{rq{gw}{cm}}{ob}{ob{mg{mg}{md}{lr{cm{lr}}{mg{gw{uk}{mg}{ob}{rq}{uk}{sn}{bz{rq}}{md{gw}{uk}{rq}{uk}{rq}{uk}{sn}{uk}{mg}{uk}{sn}}}{lr{mg}}{gw{gw}{md}{cm}{mg}{cm}{sn}{uk}{cm}{rq}}{cm{lr}{uk}{sn}{gw}}{gw}}{cm}{rq}}{rq}{mg{bz}{sn}}{sn}{gw{mg}{lr}{rq}{sn}}}{rq{mg}{sn{sn{sn}{bz}{uk}{sn}{md}{md}}}{lr{gw}{bz{gw{mg}{cm}{gw}{rq}{lr}{cm}{md{sn}{rq}{cm}}{rq}{ob{uk{mg}{cm{md}{sn}{gw{rq}{rq}{md}{lr}{ob}{rq}{ob}{sn}}{uk}{cm}{uk}{ob}{cm}{gw}{lr}}{ob}{sn}}{mg}}}{ob{sn}{rq}{sn}{md{uk}{uk}{lr}{uk}{bz}{bz{md}{gw}{ob}{md}}{ob}}{uk}{cm{rq}{cm}{ob}}}{cm{mg{lr}{md}}{sn}}}{ob}{sn}}}{mg{sn}}{md{ob}{ob}{lr{rq}{mg{md}{lr}{rq}{cm{sn}{uk}{rq}{sn}{cm}}{ob}}{uk}{lr}{rq}{cm}{lr{ob}{ob}}}{cm}{gw{cm}{uk}{ob}{gw}{cm}{sn}{lr}}{bz}{rq}{gw{gw{cm}{mg}{sn}{rq}{sn}}{cm}{gw}{md{bz}}{ob{cm}{gw}{lr}{mg}{bz}}{uk}{uk}}{bz}}{ob}{md}{lr}{gw}{gw}{mg{sn{lr}{sn}{uk}{gw}{gw}{uk}{mg}{gw}{cm}}{ob}{uk{rq}}{bz{mg}{cm}}{ob{ob}{bz}{sn}{uk}{rq}{rq}{sn}{lr}{gw}{sn}{ob{lr{lr}{mg}{md}{rq}{cm}{sn}{lr}{cm}{rq}}}{bz{rq}}{rq{rq{cm}}{sn{md}{bz}}}{cm}{cm{cm{lr}{rq}{bz}{bz}{uk}}{md}{mg{mg}{rq}{bz{uk}{uk}}{gw}}}}{rq{lr{cm}{sn}{md}{md{cm}{bz}}{ob}{ob{sn{gw}{mg}}}{ob{gw}}}}{md{lr}}{md{cm}{md}{cm{md}{uk}}{md}{rq}}}}{rq}{rq{bz{gw}}}{cm{cm{bz}{ob}}{cm{ob}}{mg{rq{gw{uk}}{md{mg}{uk}{gw}}}{ob{bz{ob}{mg}{gw{cm}{uk}{gw}{mg}{rq}}{bz}{mg{cm{mg}{ob{cm}{rq}}{mg}}{sn{sn}{mg{lr}{bz}{mg}{gw}}}{sn}}}{cm{mg{rq}{uk}}}{mg{cm}}{mg{mg}{sn{lr}{mg}}{ob{lr}{md}{md{cm{lr{lr}}{cm{uk{bz}}}{rq}{md}{ob}}}{sn{md}}}{gw}{cm}}}{bz{uk}{gw}{sn}}{uk}{ob}}{gw{mg}{cm}{uk}{cm{mg}{sn}{sn}{sn}}}{rq}{mg{md}{rq}{md{mg}{mg}}{uk}{lr{ob}{gw}{bz}}}{lr{bz}{ob}{sn}{sn}}}{mg{cm}{uk}{sn}{uk{sn{gw}{sn{rq}{rq}{bz}{rq}{rq}{bz}{mg}{rq}{cm}{gw{mg}}{uk{bz}}{gw{gw}{rq}{bz}}}{sn}{ob}{md}}{ob}{sn{lr{lr}{lr}{lr}{bz}{rq}{uk}{cm}}{lr{sn}{uk}{rq}{rq}{sn}{bz}{cm}{sn}{md}{sn}}{rq}}}{sn}}{ob}}}{lr{sn}{mg}{sn{uk}{lr}{mg}{rq}}}{bz{rq}{md{uk{bz}{ob}{md{rq}{uk{bz}{bz}{uk}{rq{mg{rq}{ob}}}{cm{cm}{ob}{mg{uk}}{bz{ob}{mg}{cm}{lr}{cm}}{lr}}{sn}{mg{md}}{rq{bz{mg}{bz}{mg}{md}{sn}}{lr{ob}{bz}{md{gw}{bz}}{lr}{md}}{sn}{cm}{gw{bz}{ob}}{sn}{ob}}{cm{cm}{mg}}}{uk{rq}}{cm{rq}}{ob{ob}{bz}{md}{md}{rq}{uk}{rq}{ob}}{mg}}{sn{sn}{ob}{rq}}}{gw}{lr}{uk}{gw}}}{ob{lr{sn}{uk}{lr}{bz}{ob}{sn}}{cm}}{mg{md}}}{cm{sn}{gw}{cm}{uk}}{rq}{uk{gw{rq{rq}{cm}{bz}{sn{lr}{rq}}{sn{sn{uk}{cm}{sn}{mg}}{md}{sn}{gw}{gw}{cm}{bz}{rq{gw}{bz}{gw}{rq}{ob}{md}}}}{md{md}{uk{lr{bz}{rq}{mg}}{sn}{rq{cm}{sn}{sn}{bz}}}{gw}{ob{cm}{mg}{bz}}{mg}{rq}{rq}{lr}{mg{uk}{bz}{gw}{mg{sn}{uk{mg}{cm}{uk}{md}{lr}}{uk{bz}{bz}{md}{sn}{sn}{lr}}{sn{bz}{rq}{lr}{gw}{cm}{gw}{gw{ob}{ob}{gw}{ob}}}}{uk}{bz}}}}{mg{mg{sn{rq}}{bz{md}}{bz{md}{sn}{uk}{lr}{sn}}{rq}}{md}{sn{cm}}{cm{lr{gw}{sn}{ob}}{gw}{mg}{ob{gw}{sn}{uk}{sn}{sn}}{md}{bz}{rq}{md{cm}{rq}{bz{uk}{mg}{sn}{mg}{rq}{cm}}}}{cm{uk{lr}{mg}{lr}{uk}{sn{cm}}{rq{sn}{uk{cm}{rq}{cm}{lr}{uk}{gw}}}{rq}{md}}}{lr{ob{sn{sn{ob}{bz}{mg}{ob{mg}{uk}{md}{bz}}{lr{sn}{cm}}{uk{lr{md}{gw}{mg}{ob}}{sn{bz}{ob}{bz{uk{bz}{mg}{rq}{lr}{sn}{lr}{sn}{rq}{mg}{gw}{uk}{gw}}{bz}}{bz}}}{lr{lr}{cm}{ob{rq}{cm}{cm}{gw}{md}{cm}{bz}}{uk}{mg}}}{sn}{gw}{uk}{rq{uk}}}{lr}{sn}{cm}{gw{gw}{sn}{gw{bz}{md}}{cm{cm}}{mg{lr}{sn}{lr}{sn{cm}{cm}{md}{md}}{lr}{gw}}}}{gw{md}{cm}{sn}}{bz}{bz{uk{bz{rq{uk}{cm}{lr}}{ob}{rq{ob}{sn}}}{uk}{rq{uk{md}{rq}{md}{gw}{mg}{ob}}{uk{mg}{uk}{rq}{mg}{bz}{sn}}{sn}}}{uk{mg}{rq}{uk}{sn}{gw}{sn}{bz}{sn{cm}{sn}{ob{cm}{bz}{rq}{mg}{ob}{rq}{md}}}}{uk{uk{gw}{cm{sn}{bz}{md}{md{md}{gw}{mg}{sn{md}}}{bz}{lr}{md}{gw}}{ob{lr}{ob}{md}{gw}{md}{ob{gw}}}{sn{cm}{lr}}{lr}{bz}{mg}{mg}}{rq}{lr{bz{bz}}{sn{gw}{lr}{ob}{sn}}{mg}{uk}{mg}{uk}{mg}}{cm{lr}{lr}{uk}{mg}{uk{ob}{rq}{md}{gw}}{bz}{lr{sn}{lr}{bz}{sn}}}{bz{mg}{ob}{sn}}}{mg}{rq{cm}{rq}{rq}{sn}{lr}{lr}{cm}}}{bz}}{sn}{cm}{mg}{ob{bz}{rq}{lr{mg}{gw}{bz}}{sn{sn}{ob}}{uk{md}{rq}{cm}{ob}{md}}{bz}{ob{rq}{gw{rq}}{uk{uk}{mg}{rq}{lr}}{ob}{uk{bz}{rq}}{rq}{mg}{cm}}{mg}{ob{bz}{rq{md{ob}{uk}}{sn}{uk}{ob{uk}{mg}{gw}}{ob}{ob}}{uk}{uk}{bz}}}}}{mg{md}{md{gw}{gw{ob{lr{cm{mg{lr}{mg}{sn}}{cm{cm}{sn}{cm}{mg}{cm}}{gw{rq}{gw}}{ob}}{uk}{ob{gw}{cm}{gw{rq{ob}}{ob{rq}{cm}{rq}{sn}{ob}{cm}}{rq{sn}}{bz{bz{ob}{ob}{lr}{ob}{cm}}{bz}{sn}{gw}{sn}{gw}{bz}{uk}}}{cm}}{cm}{md}{sn{md}{ob}{rq{sn}{gw}{uk}{gw}}{md}}{bz}{md}}{gw}{sn}{rq{rq}}{cm{gw}{rq{sn}{cm}{cm}{lr}}{mg{ob}{cm}{cm}{sn}}}{md}{lr{rq}{sn}{cm}{cm}{mg}}{uk}{gw{bz}{bz{bz}{sn}{uk}}{gw}{lr}{cm}}{cm{rq}{md}{sn{gw{md}}{ob{cm}}{ob{uk}{bz}}{gw{uk}{cm}{sn}{sn{sn}{ob}{rq{cm}{gw}{lr}}{lr{rq}{uk}{rq}{uk}{gw{sn}{sn{sn}{bz}{rq}{md}{ob}{bz}{sn{rq}{uk}{ob}}{rq}{md}{ob}}{sn}{mg}{lr}{uk}{lr}{lr}}{bz}{uk}}{ob{cm}{md}{cm}{bz}}{gw}{mg}}}{md}}{gw{rq{mg}{cm}{bz}{lr}{lr}}{uk{cm}{mg}{bz}{rq{gw{sn}{sn}{sn}}{cm}{cm}}{gw}}{bz{rq}{rq{sn}{md}{mg}{cm}{ob{sn}{uk}{cm}}}{rq}{bz{lr}{uk}}{ob{gw}{gw}{ob}{mg}{gw}{ob}}{mg}{bz{uk}{rq{rq}{mg}{bz}{bz}{gw}}}}{cm}{gw{lr{rq{md}{gw}{md}{gw}{gw}{lr}{md}{lr}}{lr{ob}}{md{md}{mg}}{lr{sn}}{lr{rq}{uk}{cm}{uk{md}{cm}{md}{lr}{sn{lr}{ob}{ob{ob}{bz}{mg}{cm}}{md}{ob{bz{lr}{bz}}{rq}{md}{sn}}{ob{sn}{md}{ob}{rq}{ob}{sn}{sn}{ob}{rq}}{ob}{rq}{rq}}{lr}{lr}}{bz}{bz{bz}{mg}}}{bz{uk}{uk{uk}{lr}{gw}{mg}{uk}}{gw}{bz}{sn}}}{uk}}{uk{lr}{gw{md}{bz}{rq}{rq}{rq{cm}{md}{mg}{sn}{gw}{mg}{md}{gw}{cm{md}}{cm{bz}{sn}{ob}{cm}}{rq}{sn}}{lr}}{gw{cm}{sn}}{rq}{gw}{sn{bz}{cm}{bz}{cm}{lr{md}{rq}{sn}{mg}}{ob}{rq{lr}}}}{md}{uk}}}{uk}}}}{md{md{ob}{md}{sn{ob{sn}{bz}{cm{lr}{rq}{mg}{md}{lr}{lr}{rq}{sn}{sn}{md}{mg}{gw}}{cm}{mg}{bz}{rq{rq{cm}{cm}{md}{mg}{cm}}{mg}{ob}}{mg}}{mg{rq{rq{md}{lr}{lr}{rq}{ob}{rq{rq{gw{rq}{sn}{mg}{gw}{lr}{lr}{ob}}}{gw}{gw}{mg}}}}{ob{mg{rq}{sn}{uk}}{rq{uk}{uk}{sn{mg}}{cm{cm}{cm}{rq}{lr{md}{uk}{sn}}}{bz}{ob}{ob}}}{md}{mg{lr}{cm}{ob}{md}}{md{ob}}}{rq}}{ob}}{rq{rq{cm{ob}{gw{cm}{uk}{lr}}{lr}}{sn}{bz{mg}{md}{bz}{bz{ob}{md}{sn}{ob}}{mg}{sn{uk}{uk}}{rq{rq}{ob}{bz}{uk}{sn}{cm}{cm}{uk}}{gw{cm}{md}}{mg}}{sn}{cm{lr}{sn}{cm{gw}{mg{ob}{ob}{ob}{gw}{lr}{lr}{cm}}{md}{lr}{bz}{uk}}{bz}{lr{ob{cm{mg}{rq}{gw}{ob{lr}{ob}}{cm{mg}{sn}}{md}}{uk{sn}{rq}{md}}{uk}{sn}{mg}{ob}}{uk}{gw{md}{cm{mg}{md}{gw}{uk}{sn}{mg}}}{rq{cm}{md}{cm}{md}{sn}{gw}{md}{cm}{uk}{uk}{sn}}{rq}{rq{uk{lr{ob}{sn}{lr}}{gw}{sn}}{ob{gw}{uk}{uk}{sn}}{uk}{rq}{sn}{ob}{md{cm}{ob}{cm{mg}{cm}{lr}}{sn}{mg}{sn{ob}{mg}{bz}{lr}}{cm}{rq{mg{uk}{sn}}{bz}{lr}{sn}{cm}{uk}{sn}{gw}{lr}{lr}{gw}{bz{cm}{sn}{lr}{sn}{ob}{lr}{uk{mg}{lr}{uk}{mg}{sn}}}{md}}}}}{cm}{uk}}{uk{sn{ob}}}{lr{sn{gw{mg}{ob}{bz}{bz{bz}}{sn}}{mg{mg}{uk}{lr}{gw}{lr}{ob}}{mg{bz}{bz}{bz}{sn}}{mg}{mg}{lr{cm}}{ob{ob}}{gw{bz}{lr}}}{uk{cm}{bz}{md}{gw}}{sn{ob}{ob}{ob}}{md{uk}}}{uk}}{lr{lr{md}}{ob{sn}{sn}{bz}{gw}}{md}{rq{md}{ob}}}{mg{sn}{mg{cm}{rq}{bz}}{gw}{md}}{gw}}{gw{uk{bz{uk}{mg}}{gw}{rq{ob}{ob}{sn}{mg}{lr}{cm}}{mg}{md}{gw}{bz}}{gw}{bz}}{ob{md}{uk}{sn}}{rq{gw{mg}{rq}}}{ob{mg{uk}{cm}{md{gw{bz}{sn}{gw}{cm}{uk{rq{cm}{rq}{sn}{mg}{cm}{md}}}{uk{sn}{uk}}{md}{rq{lr}{lr}{cm}{uk}{rq}{sn}{md}}}{lr}{rq}{lr}}{lr{lr}}{uk{rq{uk{cm}{rq}{md}{bz}{uk}{uk}{sn{rq{rq}{rq}{uk}{mg}{uk}}{bz}{md}{md}{rq}{bz{bz}{bz}{bz}{uk}{gw{bz}{lr}{md{uk}{cm}{ob}{gw}{gw}{ob}{ob}{mg}{sn}{rq}{md}}}{uk}{md}}}}}{uk{rq{ob}{bz{gw}}{ob{uk}{gw}}{sn{sn}{mg}{mg}{lr}{md}{mg}{ob}}{lr}}{lr{ob}{cm}{lr}{mg}{bz}{cm}{ob}{lr}}{md}{md}{rq{uk}{md}{gw}{lr}{md}}{gw{lr}{bz}{ob}{rq}}}{md{cm}}{sn{gw}}}}{gw{bz{rq{uk{mg}{gw}{gw}}{rq}{rq}{md}{cm}{rq}}{gw}{lr{bz}{uk}{gw}}{bz}{lr}{sn}{gw}{gw}}}{gw{rq}{sn}{gw}{gw}{mg{gw}{lr{lr}}{mg}{md}{gw}{sn}{gw{bz}{md}{sn}{uk}}}{lr}}{lr{sn}}{mg{rq{lr{uk}{md}{uk}{rq}{md}}{rq{bz}}{gw{rq}{cm}{lr}}{rq{sn}{gw{mg}{cm}}{lr}{cm}{md}}{md{mg}{cm}{lr{mg}{md}}{bz}{ob{cm}{cm}}{cm}}{bz{md}{md}{rq{md}{bz}{ob}{sn}{cm}{rq}{sn}}{mg}{ob}{uk}{mg}}}{ob{rq{bz}{md}{gw}}{bz}{sn}{lr}}{lr}}{bz{uk}}{gw{md{rq{lr{mg}{rq{cm}{uk}{cm}}{uk{lr}}{rq{lr{uk{cm}{rq}}{mg}{bz}{ob{lr}{md{cm}{ob}{uk}{ob}{uk}{md}{uk}}{mg}{gw{lr}{mg}}{lr}{lr}}{cm}{bz{md}}{uk{sn}{uk}}{sn}{gw}{rq{bz{rq}{md}{md}{cm}{mg}{md}{bz}{rq}{gw}{rq}{cm}{sn}}{gw}{lr}}{ob}}{mg{md}{gw}{cm}{md}{lr}{bz}{cm{cm}{bz}}}}{bz{bz}{md{gw}{bz{lr}{mg}{uk}{uk}}{uk{bz}{uk}{lr}}{uk{ob}}{sn}{md}{ob}{cm}{cm}{md}{sn}{bz}}{ob}{sn}{bz}{gw}{rq{mg}}}{ob}}}{uk{sn}{rq{cm}}{cm{gw{cm{lr}{uk}{mg}{rq}{rq}}}{mg{cm}}}{rq{sn{rq}{md{sn{lr}}}}{md{uk{md}{bz}}{mg{ob}}{rq{bz}{sn}{sn}{uk}{lr}{md}{rq}}{rq}{uk}{md{bz}}{bz{uk}{rq}{uk}{rq{lr}{uk}{md}{sn}{uk}{sn}}{bz}{cm}}{sn}{uk}}{bz}{mg{ob}{mg}{sn}{lr}}{cm}{rq}{uk}}}{mg}{mg{mg{bz}{rq}{lr}{sn{mg}{ob}{bz}}{cm}{sn}{lr}{mg}{cm}}}{bz{bz{lr}{uk{bz{cm}{lr}{mg}{md{uk}{mg}{lr}{mg}{cm}}{mg}{rq}{mg}{gw{cm}{cm}{gw}{gw}{lr}}{sn}}{cm}{mg{sn}{lr}{cm{ob{uk{uk}{rq}{uk}{ob}{bz}{sn}}{mg}{rq}{gw}{uk}{rq}{uk}}{bz{lr}{cm}{cm}{cm}{gw}{cm}{sn}{lr}{gw}{md}{cm}}{lr}}{md{ob{ob}{mg}}{lr}{sn}{cm{md}{gw}{sn}{cm}{mg}{rq}{ob}}{md{gw}{mg}{ob}{gw}{uk}{ob}{gw}}{mg{cm}}{uk{rq}{mg}}}}{ob}{lr}{mg{sn}{md}{md}}{lr{lr}{bz}{gw}{gw}{uk}{cm}}{mg}{ob}{md{gw}}{mg{bz}{mg}{sn}{cm}{sn}{mg}}{rq}{cm{ob{md}{rq}}}{cm{uk}{gw}{mg}{mg}{rq}{rq}}{md{mg}{sn}{bz}{gw}{sn}{gw{sn}{ob}{ob{uk}{bz}{lr}{rq}{mg}{mg}{mg}{sn}{bz}}{ob{uk}{mg}}{uk}{ob}{mg}{rq}{gw}{sn}{sn}{uk}}}}{bz}{uk}{ob}{gw{gw{uk}{bz}{md}{cm}{ob}{rq}{ob}}{md}{mg{md}{bz}{mg}{sn}}{lr}{mg}}{md{gw}{uk{lr}{bz}}}{sn{gw}}{cm{rq}{sn{lr}{rq}{mg}{cm}{lr}{sn}}{lr}{uk}{sn}{ob{uk}{uk}{sn}}}{sn}{ob}{rq{ob}{gw}{lr}{ob}{cm}{cm}}}{sn{md}{cm}{cm}}{rq}{ob}{lr{ob}{bz}{bz}}{cm{sn}{sn}{bz}{ob}{ob}{md}{rq{ob}}}{bz{gw}{lr{mg}{gw}{bz}{cm}{cm}{ob}{md{mg}{md}{cm}{mg}{bz}{md}{mg}}{gw}{cm}{ob{gw}{cm}{lr{md}}}{bz}{cm{cm}{gw}{bz}{md}{cm}{cm}{rq}}}}{gw{ob{lr}{gw}}}{bz{cm{bz}{cm{ob{lr}{mg}}{uk}{lr{bz}{cm}{rq}{sn}{cm}{bz}{bz}{mg}{md}}{cm}}{bz{lr}{md}{rq}{gw}}{ob{uk}{sn}{sn}{bz}{uk}{uk}{mg{mg}{uk}{sn}{uk}{lr}{rq}{lr}{uk}{rq}}{gw}}}{mg{mg{gw{rq}{cm}{ob{gw{lr}}{uk{uk}{gw}{sn{sn}{ob}{md}{sn}{md}{rq}}{bz}{mg{md}{rq}{sn}{cm}{md}{lr}{gw}{gw}}}}{sn{sn}{sn}{mg}{sn}{md}}}{cm}{rq}{lr}{lr{rq}}{rq{bz}{sn}{ob}}{sn}{bz}{bz}}{mg{mg{uk}{cm}{sn}{mg}}{gw{sn}{lr}}{ob}{lr}{ob{md}{lr}{lr}{lr}{cm}{lr}{sn{gw}}}{mg{gw{rq{lr}{sn}{md}}{sn{cm}{mg}{lr}{bz}{lr{mg}{bz}{gw}{uk}{bz}{bz}{lr}{bz}}{gw{sn}{ob}{ob}{uk}}{lr}{ob{lr}{bz}{uk}{gw}}}}{lr{ob}{mg}{bz{md}{bz}}{rq}{rq}}{sn{gw}{uk}{bz}{bz}{gw}{lr}{mg}{cm}{lr}{cm}}}{gw}{bz{uk}{cm}{gw}{uk}{md}{uk}{uk{lr}{cm{bz}}{mg{sn}{bz}{bz}{md}{bz}{gw}}{sn}{rq{md}{bz}{sn}{rq}{bz{uk}}{md{cm}{ob}{ob}}}{sn}{lr}{rq}{uk}}{gw}}{ob{uk}{ob}}}{md}{md{ob{cm}{uk}{lr}}{lr}{rq{md}{rq}{ob}{mg{bz}{cm}{sn{uk}{gw}}}{uk}{lr}}{gw{gw}}{md{md}{cm}{md}{sn}}{uk}{mg}{bz}{uk}}{ob{gw}}{ob{gw{ob}}}{sn{ob{mg{rq}{cm}{bz}{mg{lr}{uk}{mg}{ob}}}{gw}{uk}}}{cm{mg{bz}{rq}{bz}{lr}{gw}{uk}{ob}{mg}{ob}{sn}}{gw{cm}{gw}{ob}{rq}{mg}{rq}{uk}{rq}}{cm{cm}{uk}{md{mg{gw}{lr}}}}{rq{mg}{md{sn{bz}{md}{sn}}}{bz{md}{gw}{ob}}{lr}{cm}{rq}{lr}}{gw}{rq}{gw{mg}{bz{rq}{sn}{cm}{gw{bz}{mg}{bz}}{ob}}{sn}{rq}{lr{lr}{lr}{gw}{bz{ob}{sn}{md}}}{uk}{cm}{md{cm}{mg}{uk}{md{mg}{sn}{cm}}{bz}{gw}{cm}}{md}{mg}{uk{sn}{rq}{lr}{md}{sn}}{lr}{uk{bz{md}{gw}}{rq}{cm}{gw{uk}{ob{cm{md}{sn}{md}{rq}{rq}}{bz{rq}}{sn{md}{lr}}{uk{lr}{mg}}{bz}{md}{cm}{bz}{uk}}{cm}{md{gw}{md}{cm}}{ob}{lr}{ob}}{sn}}{gw{md}{cm}{rq}{rq}{md}{ob}{sn}{sn}{sn}}}}{uk}{md}{uk}{cm}{rq{sn}{sn}{gw}{uk}{uk}}{mg}}{rq{sn}{gw}{bz}{bz}{mg}{uk}{sn}}{gw{gw}{ob}{uk}{ob}{sn}{mg}{md}{cm}{mg{sn}{md}{cm}}{lr}{gw}}}}{cm}{uk{gw{mg}}{mg{rq{ob}{sn{ob}}{mg{mg}{uk}{cm}}{sn}{bz}{mg{md}{sn{sn}}{md{mg{mg}{uk}{gw}{lr}{md}{md}{rq}{cm}{uk}{gw}{lr}{sn}{uk}{cm}}{bz}{lr}}{uk}{cm}{sn}{md}}{bz}{uk}{sn}{md{bz}{rq}{gw{uk}{md}{bz}{rq}}}{cm{cm}{gw}{sn}{gw}{mg}{uk}{bz}{bz}{sn}{bz}}}{uk}{mg}{rq{rq{lr{gw}{sn}{ob}{rq{gw}{uk}{gw{ob{ob}}{sn{sn}}{lr{md{rq}{sn}{sn}{bz}}{md}}{gw}{gw{mg}{mg}{lr}{gw}{bz}}{rq{ob}{ob}{cm}}{mg}}{uk}{ob}{md}}}{gw{uk{gw}{mg}{gw}{uk}{rq}}{uk}{lr{mg}{bz}{uk}{lr}{mg}{sn}{lr}}{gw}{ob}{md}{md}}{mg}{mg}{mg}{uk}}}{mg{mg}{rq}{gw}{bz{uk}{md{sn}{bz}{gw}{md}{lr}{gw}{cm{gw}{lr}{cm}}{rq}{sn}{ob}{bz}{sn}{cm}}}{ob{bz{uk}{bz}}{cm{cm}{sn}{bz}{ob}{gw}{ob}}{rq}{uk{lr{lr}{gw}{uk}{uk}{lr{gw}}{cm{mg}{mg}{gw}{cm}{sn}}{gw}}{ob}}}{gw}{uk{cm}{mg}{lr}{bz}{ob}{sn{rq}}}{mg{cm}{bz}{md}{sn}}}}{cm{sn}{mg{rq}}}{lr}{bz{uk{mg}{rq}{gw}{gw}{lr}{lr}{lr}{gw}{lr}{bz{mg}{rq}{gw}{uk}{cm}{rq}}{uk}}{mg{rq}{cm}{md}{lr{rq}{bz}{uk}{gw}{uk}{bz}}{bz}{bz}{sn}{cm}}{sn}}}{cm}{lr{mg}{lr{md}{mg}{rq{ob}{lr}}}{bz}}}{cm{sn}{bz}{mg}{cm}{md}}{mg{mg}{sn{sn}{bz{bz}{sn{bz{rq}{mg}{gw{ob}{gw{lr}{gw{gw}{mg}{lr}}{gw}{sn}{lr}{cm}{gw}{lr}}{mg}{bz{uk}{sn}{gw}{ob}}{uk}{uk}{bz}{lr}}}{bz{ob}{ob}}{cm}{mg}{mg}{mg}{lr}}}{mg}{gw}}{bz}{gw{mg}}}}{gw{cm{rq}{cm{uk{bz}{ob}{lr}{rq}{uk}{ob}{cm}}{rq{lr}{uk}}{ob}{sn}{bz}{lr{uk}{mg}{sn}}{mg}{rq}{mg{uk}}}{lr}{uk}{lr}}{ob}{rq{sn}{bz{md}{ob}{md}{cm}{sn}{uk}}{cm}{md}{uk}{md}}}{ob{ob{gw{mg{mg{gw}{rq}{gw}{ob{uk}{ob}{mg{uk}{rq}{md}{mg}}{sn}}}{sn{ob}{bz{md{sn}{md}{cm}}}{cm{lr}{md}}{md}}{sn}{rq}}{lr}{ob{rq}{ob}{md}{gw}{bz}}{mg{ob}{mg}{md}{cm}{bz}{rq}}{rq}{uk}{ob}{cm}}}{gw}{bz}{uk}{gw{md}{lr{ob}{md}}{bz{cm}{uk{gw}}}{cm}}{sn}{sn}{md{sn}{mg}}{mg{gw}{lr}{mg}{lr}{ob}{lr}{uk}}{mg{lr}{sn}{rq}{sn}{rq{sn}}{cm{rq}{rq}{gw}{gw}{mg}{bz{mg}{ob}{md}{md}{rq}}{sn{md}{uk}{cm}}{lr}{ob}}{uk}}}{gw{ob}{md}{ob{ob}{uk}{gw}{bz}{rq}}{rq}{sn{ob}{uk}{uk}{bz}}{lr}}}{rq{sn}{uk{uk}{bz{rq}{cm}}{mg}{cm}{rq{sn}{bz}{lr}{md}}{md}{bz}}{lr}{sn}{mg{md}{sn}{uk}{md}{uk}{cm}{sn}{mg}}{lr{md}{md}{rq}{gw}}}}}{ob}{cm}{uk{ob{uk{bz}{sn{rq{rq}{uk}{bz}{cm}{lr{lr}{cm}{gw{mg}}{sn{md}}{bz{bz}{lr}{ob}{uk}}{md{rq}{rq}}}}{gw{sn}}}{mg{mg{gw}{cm}{gw}{md}{uk{uk{gw}{bz}{ob}{uk}{gw}}{gw}{md}{cm}}{uk}{gw}{ob}}{lr{md{sn}{cm{ob}{uk}{rq}}{cm{mg}{mg}}{lr{cm{mg{md}{ob}{ob}{gw{cm{md}}{cm{uk}{mg}{ob}}{md}{mg}{bz}{uk}{lr}{lr}}{cm{cm}{ob}}{uk}{ob}{uk}}{gw{uk}{sn}}{lr{bz}{lr}{cm}}}}{bz{gw}{uk}}{rq{mg}{lr{sn}{cm}{rq}}}{mg{mg{uk{mg}{sn}{cm}}{lr}{gw}{sn}{lr{md}{sn{ob{uk}{bz}}{cm{md{cm}{lr}}{md}{md}{mg}}{uk}{gw{lr}{bz}{uk{md}{bz}{uk}{mg}{rq{gw}{bz}{lr}{ob}{ob}{cm}}{gw{gw}}}}}{gw}{sn}}}{md}{md}}}{bz}}}{gw{mg}{md}}{rq}}{ob{sn}{mg}{mg}{sn}{gw}{mg{uk}{md}{cm}}}{ob}{md}}{mg{md{gw}}{cm{cm}{lr{bz}}{sn{ob}{lr}{rq}{uk}{uk}{lr}}{md}{sn}{md{rq}{bz}{rq}{mg}{rq{rq}}}}{gw{gw}{sn}}}}}{uk}}");
		System.out.println("parent: " + t);
		LblTree child = (LblTree)t.getChildAt(0);
		child.removeFromParent();
		System.out.println("child: " + child);
		LblTree granchild = (LblTree)child.getChildAt(0);
		System.out.println("granchild: " + granchild);
		t.insert(granchild, 0);
		System.out.println("Inserting granchild to parent...");
		System.out.println("parent: " + t);
		System.out.println("child: " + child);
		System.out.println("granchild: " + granchild);
	//	AdjacencyEncForest f=new AdjacencyEncForest();
			
	}

	public int getNodeID() {
		return nodeID;
	}

	public void setNodeID(int nodeID) {
		this.nodeID = nodeID;
	}
	
}
